# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Employ DMET as a problem decomposition technique."""

from enum import Enum
from functools import reduce
import numpy as np
from pyscf import gto, scf
import scipy
import warnings

from tangelo.problem_decomposition.dmet import _helpers as helpers
from tangelo.problem_decomposition.problem_decomposition import ProblemDecomposition
from tangelo.problem_decomposition.electron_localization import iao_localization, meta_lowdin_localization
from tangelo.problem_decomposition.dmet.fragment import SecondQuantizedDMETFragment
from tangelo.algorithms import FCISolver, CCSDSolver, VQESolver
from tangelo.toolboxes.post_processing.mc_weeny_rdm_purification import mcweeny_purify_2rdm


class Localization(Enum):
    """Enumeration of the electron localization supported by DMET."""
    meta_lowdin = 0
    iao = 1


class DMETProblemDecomposition(ProblemDecomposition):
    """DMET single-shot algorithm is used for problem decomposition technique.
    By default, CCSD is used as the electronic structure solver, and Meta-Lowdin
    is used for the localization scheme. Users can define other electronic
    structure solver such as FCI or VQE as an impurity solver. IAO scheme can be
    used instead of the Meta-Lowdin localization scheme, but it cannot be used
    for minimal basis set.

    Attributes:
        molecule (SecondQuantizedMolecule): The molecular system.
        electron_localization (Localization): A type of localization scheme.
            Default is Meta-Lowdin.
        fragment_atoms (list): List of number of atoms in each fragment. Sum of
            this list should be the same as the number of atoms in the original
            system.
        fragment_solvers (list or str): List of solvers for each fragment. If
            only a string is detected, this solver is used for all fragments.
        optimizer (function handle): A function defining the classical optimizer
            and its behavior.
        initial_chemical_potential (float) : Initial value for the chemical
            potential.
        solvers_options (list or dict): List of dictionaries for the solver
            options. If only a single dictionary is passed, the same options are
            applied for every solver. This will raise an error if different
            solvers are parsed.
        verbose (bool) : Flag for DMET verbosity.
    """

    def __init__(self, opt_dict):

        default_ccsd_options = dict()
        default_fci_options = dict()
        default_vqe_options = {"qubit_mapping": "jw",
                               "initial_var_params": "ones",
                               "verbose": False}

        default_options = {"molecule": None,
                           "electron_localization": Localization.meta_lowdin,
                           "fragment_atoms": list(),
                           "fragment_solvers": "ccsd",
                           "optimizer": self._default_optimizer,
                           "initial_chemical_potential": 0.0,
                           "solvers_options": list(),
                           "verbose": False}

        self.builtin_localization = set(Localization)

        # Initialize with default values
        self.__dict__ = default_options
        # Overwrite default values with user-provided ones, if they correspond to a valid keyword
        for k, v in opt_dict.items():
            if k in default_options:
                setattr(self, k, v)
            else:
                raise KeyError(f"Keyword :: {k}, not available in DMETProblemDecomposition.")

        # Raise error/warnings if input is not as expected
        if not self.molecule:
            raise ValueError(f"A SecondQuantizedMolecule object must be provided when instantiating DMETProblemDecomposition.")

        # Converting our interface to pyscf.mol.gto and pyscf.scf (used by this
        # code).
        self.mean_field = self.molecule.mean_field
        self.molecule = self.molecule.to_pyscf(self.molecule.basis)

        # If fragment_atoms is detected as a nested list of int, atoms are reordered to be
        # consistent with a list of numbers representing the number of atoms in each fragment.
        if isinstance(self.fragment_atoms, list) and all(isinstance(list_atoms, list) for list_atoms in self.fragment_atoms):
            fragment_atoms_flatten = [atom_id for frag in self.fragment_atoms for atom_id in frag]

            if max(fragment_atoms_flatten) >= self.molecule.natm:
                raise RuntimeError("An atom id is higher than the number of atom (indices start at 0).")
            elif len(fragment_atoms_flatten) != len(set(fragment_atoms_flatten)):
                raise RuntimeError("Atom indices must only appear once.")

            # Converting fragment_atoms to an expected list of number of atoms (not atom ids).
            new_fragment_atoms = [len(frag) for frag in self.fragment_atoms]

            # Reordering the molecule geometry.
            new_geometry = [self.molecule._atom[atom_id] for atom_id in fragment_atoms_flatten]

            # Building a new PySCF molecule with correct ordering.
            new_molecule = gto.Mole()
            new_molecule.atom = new_geometry
            new_molecule.basis = self.molecule.basis
            new_molecule.charge = self.molecule.charge
            new_molecule.spin = self.molecule.spin
            new_molecule.unit = "B"
            new_molecule.build()

            # Attribution of the expected fragment_atoms and a reordered molecule.
            self.molecule = new_molecule
            self.fragment_atoms = new_fragment_atoms

            # Force recomputing the mean field if the atom ordering has been changed.
            warnings.warn("The mean field will be recomputed even if one has been provided by the user.", RuntimeWarning)
            self.mean_field = scf.RHF(self.molecule)
            self.mean_field.verbose = 0
            self.mean_field.scf()

        # Check if the number of fragment sites is equal to the number of atoms in the molecule
        if self.molecule.natm != sum(self.fragment_atoms):
            raise RuntimeError("The number of fragment sites is not equal to the number of atoms in the molecule")

        # Check that the number of solvers matches the number of fragments.
        # If a single string is detected, it is converted to a list.
        # If a list is detected, it must have the same length than the fragment_atoms.
        if isinstance(self.fragment_solvers, str):
            self.fragment_solvers = [self.fragment_solvers for _ in range(len(self.fragment_atoms))]
        elif isinstance(self.fragment_solvers, list):
            if len(self.fragment_solvers) != len(self.fragment_atoms):
                raise RuntimeError("The number of solvers does not match the number of fragments.")

        # Check that the number of solvers options matches the number of solvers.
        # If there is no options, all default ones are applied.
        # If a single options dictionary is parsed, it is repeated for every solvers.
        # If it is a list, we verified that the length is the same.
        if not self.solvers_options:
            for solver in self.fragment_solvers:
                if solver == "ccsd":
                    self.solvers_options.append(default_ccsd_options)
                elif solver == "fci":
                    self.solvers_options.append(default_fci_options)
                elif solver == "vqe":
                    self.solvers_options.append(default_vqe_options)
        elif isinstance(self.solvers_options, dict):
            self.solvers_options = [self.solvers_options for _ in self.fragment_solvers]
        elif isinstance(self.solvers_options, list):
            if len(self.solvers_options) != len(self.fragment_solvers):
                raise RuntimeError("The number of solvers options does not match the number of solvers.")

        # Results of the DMET loops.
        self.chemical_potential = None
        self.dmet_energy = None

        # Define during the building phase (self.build()).
        self.orbitals = None
        self.orb_list = None
        self.orb_list2 = None
        self.onerdm_low = None

    @property
    def quantum_fragments_data(self):
        """This aims to return a dictionary with all necessary components to
        run a quantum experiment for a (or more) DMET fragment(s).
        """

        if not self.solver_fragment_dict:
            raise RuntimeError("Simulate method must be called beforehand.")

        quantum_data = dict()

        # Construction of a dict with SecondQuantizedDMETFragment, qubit
        # hamiltonian and quantum circuit.
        for fragment_i, vqe_object in self.solver_fragment_dict.items():
            quantum_data[fragment_i] = (vqe_object.molecule,
                vqe_object.qubit_hamiltonian, vqe_object.optimal_circuit)

        return quantum_data

    def build(self):
        """Building the orbitals list for each fragment. It sets the values of
        self.orbitals, self.orb_list and self.orb_list2.
        """

        # Locate electron with a built-in method. A custom one can be provided.
        if type(self.electron_localization) == Localization:
            if self.electron_localization == Localization.meta_lowdin:
                self.electron_localization = meta_lowdin_localization
            elif self.electron_localization == Localization.iao:
                self.electron_localization = iao_localization
            else:
                raise ValueError(f"Unsupported ansatz. Built-in localization methods:\n\t{self.builtin_localization}")
        elif not callable(self.electron_localization):
            raise TypeError(f"Invalid electron localization function. Expecting a function.")

        # Construct orbital object.
        self.orbitals = helpers._orbitals(self.molecule, self.mean_field, range(self.molecule.nao_nr()), self.electron_localization)

        # TODO: remove last argument, combining fragments not supported.
        self.orb_list, self.orb_list2, _ = helpers._fragment_constructor(self.molecule, self.fragment_atoms, 0)

        # Calculate the 1-RDM for the entire molecule.
        self.onerdm_low = helpers._low_rdm(self.orbitals.active_fock, self.orbitals.number_active_electrons)

    def simulate(self):
        """Perform DMET loop to optimize the chemical potential. It converges
        when the electron summation across all fragments is the same as the
        number of electron in the molecule.

        Returns:
            float: The DMET energy (dmet_energy).
        """

        # To keep track the number of iteration (was done with an energy list before).
        # TODO: A decorator function to do the same thing?
        self.n_iter = 0

        # Initialize the energy list and SCF procedure employing newton-raphson algorithm.
        # TODO : find a better initial guess than 0.0 for chemical potential. DMET fails often currently.
        if not self.orbitals:
            raise RuntimeError("No fragment built. Have you called DMET.build ?")

        self.chemical_potential = self.optimizer(self._oneshot_loop, self.initial_chemical_potential)
        self.chemical_potential = self.chemical_potential.real

        # run one more time to save results
        _ = self._oneshot_loop(self.chemical_potential, save_results=True)

        if self.verbose:
            print(" \t*** DMET Cycle Done *** ")
            print(" \tDMET Energy ( a.u. ) = " + "{:17.10f}".format(self.dmet_energy))
            print(" \tChemical Potential   = " + "{:17.10f}".format(self.chemical_potential))

        return self.dmet_energy

    def energy_error_bars(self, n_shots, n_resamples, purify=False, rdm_measurements=None):
        """Perform bootstrapping of measured qubit terms in RDMs to obtain error
        bars for the calculated energy. Can also perform McWeeny purification of
        the RDMs between resampling and calculating the energy.

        Args:
            n_shots (int): The number of shots used to resample from qubit
                terms.
            n_resamples (int): The number of bootstrapping samples for the
                estimate of the energy and standard deviation.
            purify (bool): Use mc_weeny_rdm_purification technique on 2-RDMs.
                Will only apply to fragments with 2 electrons.
            rdm_measurements (dict): A dictionary with keys being the DMET
                fragment number and corresponding values of a dictionary with
                keys of qubit terms in the rdm and corresponding values of a
                frequencies dictionary of the measurements.
                Example: {
                0: {((0, "X"), (1, "Y")): {"10": 0.5, "01": 0.5},
                    ((1, "Z"), (1, "X")): {"00": 0.25, "11": 0.25, "01": 0.5}}
                1: {((0, "Z")): {"0000": 1}
                    ((0, "X"), (1, "Y")): {"1111": 0.5, "0101": 0.5}}
                }.
                If run _oneshot_loop with save_results=True is called:
                self.rdm_measurements will have the proper dictionary format
                with all values populated.

        Returns:
            float: The bootstrapped DMET energy and standard deviation.
        """

        if self.chemical_potential is None:
            raise RuntimeError("No chemical_potential. Have you run a simulation yet?")

        # begin resampling
        resampled_energies = np.zeros(n_resamples, dtype=float)
        for i in range(n_resamples):
            _ = self._oneshot_loop(self.chemical_potential, save_results=False, resample=True,
                                   n_shots=n_shots, purify=purify, rdm_measurements=rdm_measurements)
            resampled_energies[i] = self.dmet_energy.real

        energy_average, energy_standard_deviation = np.mean(resampled_energies), np.std(resampled_energies, ddof=1)

        return energy_average, energy_standard_deviation

    def _build_scf_fragments(self, chemical_potential):
        """Building the orbitals list for each fragment. It sets the values of
        self.orbitals, self.orb_list and self.orb_list2.

        Args:
            onerdm_low (matrix): 1-RDM for the whole molecule.
            chemical_potential (float): Variational parameter for DMET.

        Returns:
            list: List of many objects important for each fragments.
        """

        # Empty list, all fragment informations will be returned in this list.
        scf_fragments = list()

        for i, norb in enumerate(self.orb_list):
            t_list = list()
            t_list.append(norb)
            temp_list = self.orb_list2[i]

            # Construct bath orbitals.
            bath_orb, e_occupied = helpers._fragment_bath(self.orbitals.mol_full, t_list, temp_list, self.onerdm_low)

            # Obtain one particle rdm for a fragment.
            norb_high, nelec_high, onerdm_high = helpers._fragment_rdm(t_list, bath_orb, e_occupied,
                                                                       self.orbitals.number_active_electrons)

            # Obtain one particle rdm for a fragment.
            one_ele, fock, two_ele = self.orbitals.dmet_fragment_hamiltonian(bath_orb, norb_high, onerdm_high)

            # Construct guess orbitals for fragment SCF calculations.
            guess_orbitals = helpers._fragment_guess(t_list, bath_orb, chemical_potential, norb_high, nelec_high,
                                                     self.orbitals.active_fock)

            # Carry out SCF calculation for a fragment.
            mf_fragment, fock_frag_copy, mol_frag = helpers._fragment_scf(t_list, two_ele, fock, nelec_high, norb_high,
                                                                          guess_orbitals, chemical_potential)

            scf_fragments.append([mf_fragment, fock_frag_copy, mol_frag, t_list, one_ele, two_ele, fock])

        return scf_fragments

    def _oneshot_loop(self, chemical_potential, save_results=False, resample=False, n_shots=None, purify=False, rdm_measurements=None):
        """Perform the DMET loop. This is the cost function which is optimized
        with respect to the chemical potential.

        Args:
            chemical_potential (float): The chemical potential.
            save_results (bool): If True, the VQESolver Class (solver_fragment)
                for each fragment that uses VQE is saved in
                self.solver_fragment_dict with the key being its fragment
                number.
            resample (bool): If True, the saved frequencies are resampled using
                bootstrapping.
            n_shots (int): The number of shots used for resampling, Ideally the
                same as used in the simulation but can be different.
            purify (bool): If True, use McWeeny"s purification technique to
                purify 2-RDM. Only called for fragments with 2 electrons.
            rdm_measurements (dict): A dictionary with keys being the DMET
                fragment number and corresponding values of a dictionary with
                keys of qubit terms in the rdm and corresponding values of a
                frequencies dictionary of the measurements.
                Example: {
                0: {((0, "X"), (1, "Y")): {"10": 0.5, "01": 0.5},
                    ((1, "Z"), (1, "X")): {"00": 0.25, "11": 0.25, "01": 0.5}}
                1: {((0, "Z")): {"0000": 1}
                    ((0, "X"), (1, "Y")): {"1111": 0.5, "0101": 0.5}}
                }.
                If run _oneshot_loop with save_results=True is called:
                self.rdm_measurements will have the proper dictionary format
                with all values populated.

        Returns:
            float: The new chemical potential.
        """

        self.n_iter += 1
        if self.verbose:
            print(" \tIteration = ", self.n_iter)
            print(" \t----------------")
            print(" ")

        number_of_electron = 0.0
        energy_temp = 0.0

        # Possibly add dictionary of measured frequencies for each fragment
        if resample:
            if save_results:
                raise ValueError("Can not save results and resample in same run. Must run saveresults first")
            if not hasattr(self, "solver_fragment_dict"):
                raise AttributeError("Need to run _oneshot_loop with save_results=True in order to resample")
            if rdm_measurements:
                for k, v in rdm_measurements.items():
                    self.solver_fragment_dict[k].rdm_freq_dict = v
            scf_fragments = self.scf_fragments
        else:
            # Carry out SCF calculation for all fragments.
            scf_fragments = self._build_scf_fragments(chemical_potential.real)
        if save_results:
            self.solver_fragment_dict = dict()
            self.rdm_measurements = dict()

        # Iterate across all fragment and compute their energies.
        # The total energy is stored in energy_temp.
        for i, info_fragment in enumerate(scf_fragments):

            # Unpacking the information for the selected fragment.
            mf_fragment, fock_frag_copy, mol_frag, t_list, one_ele, two_ele, fock = info_fragment

            # Interface with our data strcuture.
            # We create a dummy SecondQuantizedMolecule with a DMETFragment class.
            # It has the same important attributes and methods to be used with
            # functions of this package.
            dummy_mol = SecondQuantizedDMETFragment(mol_frag, mf_fragment, fock, fock_frag_copy, t_list, one_ele, two_ele)

            if self.verbose:
                print("\t\tFragment Number : # ", i + 1)
                print("\t\t------------------------")

            # TODO: Changing this into something more simple is preferable. There
            # would be an enum class with every solver in it. After this, we would
            # define every solver in a list and call them recursively.
            # FCISolver and CCSDSolver must be taken care of, but this is a PR itself.
            solver_fragment = self.fragment_solvers[i]
            solver_options = self.solvers_options[i]
            if solver_fragment == "fci":
                solver_fragment = FCISolver(dummy_mol, **solver_options)
                solver_fragment.simulate()
                onerdm, twordm = solver_fragment.get_rdm()
            elif solver_fragment == "ccsd":
                solver_fragment = CCSDSolver(dummy_mol, **solver_options)
                solver_fragment.simulate()
                onerdm, twordm = solver_fragment.get_rdm()
            elif solver_fragment == "vqe":
                if resample:
                    solver_fragment = self.solver_fragment_dict[i]
                    if rdm_measurements:
                        if i in rdm_measurements:
                            solver_fragment.rdm_freq_dict = rdm_measurements[i]
                        else:
                            raise KeyError(f"rdm_measurements for fragment {i} are missing")
                    if n_shots:
                        solver_fragment.backend.n_shots = n_shots
                    if solver_fragment.backend.n_shots is None:
                        raise ValueError("n_shots must be specified in original calculation or in error calculation")
                else:
                    system = {"molecule": dummy_mol}
                    solver_fragment = VQESolver({**system, **solver_options})
                    solver_fragment.build()
                    solver_fragment.simulate()

                if purify and solver_fragment.molecule.n_active_electrons == 2:
                    onerdm, twordm = solver_fragment.get_rdm(solver_fragment.optimal_var_params, resample=resample, sum_spin=False)
                    onerdm, twordm = mcweeny_purify_2rdm(twordm)
                else:
                    onerdm, twordm = solver_fragment.get_rdm(solver_fragment.optimal_var_params, resample=resample)
                if save_results:
                    self.solver_fragment_dict[i] = solver_fragment
                    self.rdm_measurements[i] = self.solver_fragment_dict[i].rdm_freq_dict

            fragment_energy, _, one_rdm = self._compute_energy(mf_fragment, onerdm, twordm,
                                                               fock_frag_copy, t_list, one_ele,
                                                               two_ele, fock)

            # Sum up the energy.
            energy_temp += fragment_energy

            # Sum up the number of electrons.
            number_of_electron += np.trace(one_rdm[: t_list[0], : t_list[0]])

            if self.verbose:
                print("\t\tFragment Energy                 = " + "{:17.10f}".format(fragment_energy))
                print("\t\tNumber of Electrons in Fragment = " + "{:17.10f}".format(np.trace(one_rdm)))
                print("")

        energy_temp += self.orbitals.core_constant_energy
        self.dmet_energy = energy_temp.real

        if save_results:
            self.scf_fragments = scf_fragments

        return number_of_electron - self.orbitals.number_active_electrons

    def get_resources(self):
        """Estimate the resources required by DMET. Only supports fragments
        solved with VQESolver. Resources for each fragments are outputed as a
        list.
        """

        # Carry out SCF calculation for all fragments.
        scf_fragments = self._build_scf_fragments(self.initial_chemical_potential)

        # Store resources for each fragments.
        resources_fragments = dict()

        # Iterate across all fragment and compute their energies.
        # The total energy is stored in energy_temp.
        for i, info_fragment in enumerate(scf_fragments):

            # Unpacking the information for the selected fragment.
            mf_fragment, fock_frag_copy, mol_frag, t_list, one_ele, two_ele, fock = info_fragment

            dummy_mol = SecondQuantizedDMETFragment(mol_frag, mf_fragment, fock, fock_frag_copy, t_list, one_ele, two_ele)

            # Buiding SCF fragments and quantum circuit. Resources are then
            # estimated. For classical sovlers, this functionality is not
            # implemented yet.
            solver_fragment = self.fragment_solvers[i]
            solver_options = self.solvers_options[i]
            if solver_fragment == "vqe":
                system = {"molecule": dummy_mol}
                solver_fragment = VQESolver({**system, **solver_options})
                solver_fragment.build()
                vqe_resources = solver_fragment.get_resources()
                resources_fragments[i] = vqe_resources

                if self.verbose:
                    print("\t\tFragment Number : # ", i + 1)
                    print("\t\t------------------------")
                    print(f"\t\t{vqe_resources}\n")

        return resources_fragments

    def _compute_energy(self, mf_frag, onerdm, twordm, fock_frag_copy, t_list, oneint, twoint, fock):
        """Calculate the fragment energy.

        Args:
            mf_frag (pyscf.scf.RHF): The mean field of the fragment.
            onerdm (numpy.array): one-particle reduced density matrix (float).
            twordm (numpy.array): two-particle reduced density matrix (float).
            fock_frag_copy (numpy.array): Fock matrix with the chemical
                potential subtracted (float).
            t_list (list): List of number of fragment and bath orbitals (int).
            oneint (numpy.array): One-electron integrals of fragment (float).
            twoint (numpy.array): Two-electron integrals of fragment (float).
            fock (numpy.array): Fock matrix of fragment (float).

        Returns:
            float: Fragment energy (fragment_energy).
            float: Total energy for fragment using RDMs (total_energy_rdm).
            numpy.array: One-particle RDM for a fragment (one_rdm, float).
        """

        # Execute CCSD calculation
        norb = t_list[0]

        # Calculate the one- and two- RDM for DMET energy calculation (Transform to AO basis)
        one_rdm = reduce(np.dot, (mf_frag.mo_coeff, onerdm, mf_frag.mo_coeff.T))

        twordm = np.einsum("pi,ijkl->pjkl", mf_frag.mo_coeff, twordm)
        twordm = np.einsum("qj,pjkl->pqkl", mf_frag.mo_coeff, twordm)
        twordm = np.einsum("rk,pqkl->pqrl", mf_frag.mo_coeff, twordm)
        twordm = np.einsum("sl,pqrl->pqrs", mf_frag.mo_coeff, twordm)

        # Calculate the total energy based on RDMs
        total_energy_rdm = np.einsum("ij,ij->", fock_frag_copy, one_rdm) + 0.5 * np.einsum("ijkl,ijkl->", twoint,
                                                                                           twordm)

        # Calculate fragment expectation value
        fragment_energy_one_rdm = 0.25 * np.einsum("ij,ij->", one_rdm[: norb, :], fock[: norb, :] + oneint[: norb, :]) \
                                  + 0.25 * np.einsum("ij,ij->", one_rdm[:, : norb], fock[:, : norb] + oneint[:, : norb])

        fragment_energy_twordm = 0.125 * np.einsum("ijkl,ijkl->", twordm[: norb, :, :, :], twoint[: norb, :, :, :]) \
                                 + 0.125 * np.einsum("ijkl,ijkl->", twordm[:, : norb, :, :], twoint[:, : norb, :, :]) \
                                 + 0.125 * np.einsum("ijkl,ijkl->", twordm[:, :, : norb, :], twoint[:, :, : norb, :]) \
                                 + 0.125 * np.einsum("ijkl,ijkl->", twordm[:, :, :, : norb], twoint[:, :, :, : norb])

        fragment_energy = fragment_energy_one_rdm + fragment_energy_twordm

        return fragment_energy, total_energy_rdm, one_rdm

    def _default_optimizer(self, func, var_params):
        """Function used as a default optimizer for DMET when user does not
        provide one.

        Args:
            func (function handle): The function that performs energy
                estimation. This function takes var_params as input and returns
                a float.
            var_params (list): The variational parameters (float).

        Returns:
            float: The chemical potential found by the optimizer.
        """

        result = scipy.optimize.newton(func, var_params, tol=1e-5)

        return result.real



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import os
from openfermion import load_operator

from tangelo.helpers.utils import default_simulator
from tangelo.linq import translator, Simulator, Circuit
from tangelo.linq.helpers import measurement_basis_gates
from tangelo.toolboxes.operators import QubitOperator
from tangelo.toolboxes.measurements import get_measurement_estimate

path_data = os.path.dirname(os.path.abspath(__file__)) + '/data'

op1 = 1. * QubitOperator('X1 Y0')
op2 = 0.01 * QubitOperator('Z0 X1')


def assert_dict_almost_equal(d1, d2, atol):
    """ Utility function to check whether two dictionaries are almost equal, for arbitrary tolerance """
    if d1.keys() != d2.keys():
        raise AssertionError("Dictionary keys differ. Frequency dictionaries are not almost equal.\n"
                             f"d1 keys: {d1.keys()} \nd2 keys: {d2.keys()}")
    else:
        for k in d1.keys():
            if abs(d1[k] - d2[k]) > atol:
                raise AssertionError(f"Dictionary entries beyond tolerance {atol}: \n{d1} \n{d2}")
    return True


class MeasurementsTest(unittest.TestCase):

    def test_measurement_uniform(self):
        """ Test uniform measurement estimation method on simple cases """

        # Integer log10 coefficients
        mes1 = get_measurement_estimate(op1)
        mes2 = get_measurement_estimate(op2)
        mes3 = get_measurement_estimate(op2 + 0.1*op1)

        # Float log10 coefficients
        mes4 = get_measurement_estimate(0.6*op1)

        # Coefficient smaller than desired accuracy -> 0
        mes5 = get_measurement_estimate(0.0001*op1)

        # Borderline case: Instead of 1 measurement, take 100.
        mes6 = get_measurement_estimate(0.001*op1)

        # Adjust digits
        mes7 = get_measurement_estimate(op1, digits=2)

        assert_dict_almost_equal(mes1, {((0, 'Y'), (1, 'X')): 100000000}, 1)
        assert_dict_almost_equal(mes2, {((0, 'Z'), (1, 'X')): 10000}, 1)
        assert_dict_almost_equal(mes3, {((0, 'Z'), (1, 'X')): 10000, ((0, 'Y'), (1, 'X')): 1000000}, 1)
        assert_dict_almost_equal(mes4, {((0, 'Y'), (1, 'X')): 36000000}, 1)
        assert_dict_almost_equal(mes5, {((0, 'Y'), (1, 'X')): 0}, 1)
        assert_dict_almost_equal(mes6, {((0, 'Y'), (1, 'X')): 100}, 1)
        assert_dict_almost_equal(mes7, {((0, 'Y'), (1, 'X')): 1000000}, 1)

    def test_measurement_uniform_H2(self):
        """ Test on UCCSD H2 usecase that uniform measurement estimation method guarantees on average the level
        of accuracy expected by the user """

        # Load state preparation circuit
        with open(f"{path_data}/H2_UCCSD.qasm", "r") as f:
            openqasm_circ = f.read()
        abs_circ = translator._translate_openqasm2abs(openqasm_circ)

        # Load qubit Hamiltonian
        qb_ham = load_operator("mol_H2_qubitham.data", data_directory=path_data, plain_text=True)

        # Get exact expectation value using a simulator
        sim_exact = Simulator()
        freqs_exact, _ = sim_exact.simulate(abs_circ)
        exp_val_exact = sim_exact.get_expectation_value(qb_ham, abs_circ)

        # Get measurement estimate using "uniform" method
        mes_dict = get_measurement_estimate(qb_ham, digits=2)

        # Get samples and compute expectation value from measurements, using exact state preparation. Compute error.
        diffs = []
        n_repeat = 50
        for i in range(n_repeat):
            exp_val = 0.
            sim = Simulator(target=default_simulator, n_shots=1)
            for m_basis, coef in qb_ham.terms.items():
                if m_basis:
                    term_circ = abs_circ + Circuit(measurement_basis_gates(m_basis))
                    sim.n_shots = mes_dict[m_basis]
                    freqs, _ = sim.simulate(term_circ)
                    exp_val += sim.get_expectation_value_from_frequencies_oneterm(m_basis, freqs) * coef
                else:
                    exp_val += coef
            diffs.append(abs(exp_val - exp_val_exact))

        # Check that on average, we deliver the expected accuracy
        average_diff = sum(diffs)/n_repeat
        assert(average_diff <= 1e-2)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import numpy as np

from tangelo.linq import Gate


class TestGates(unittest.TestCase):
    """
        A test class to check that the circuit class is working properly, and processing the information
        of the gate objects it holds as expected.
    """

    def test_some_gates(self):
        """ Test that some basic gates can be invoked with a few different parameters, and that this information
        is printed as expected by the built-in __str__ method """

        # Create a Hadamard gate acting on qubit 2
        H_gate = Gate("H", 2)
        # Create a CNOT gate with control qubit 0 and target qubit 1
        CNOT_gate = Gate("CNOT", 1, 0)
        # Create a parameterized rotation on qubit 1 with angle 2 radians
        RX_gate = Gate("RX", 1, parameter=2.)
        # Create a parameterized rotation on qubit 1 , with an undefined angle, that will be variational
        RZ_gate = Gate("RZ", 1, parameter="an expression", is_variational=True)
        # Create a multi-controlled X gate with a numpy array
        CCCX_gate = Gate("CX", 0, control=np.array([1, 2, 4], dtype=np.int32))

        for gate in [H_gate, CNOT_gate, RX_gate, RZ_gate, CCCX_gate]:
            print(gate)

    def test_some_gates_inverse(self):
        """ Test that some basic gates can be inverted with a few different parameters, and fails when non-invertible
        parameters are passed"""

        # Create a Hadamard gate acting on qubit 2
        H_gate = Gate("H", 2)
        H_gate_inverse = Gate("H", 2)
        self.assertEqual(H_gate.inverse().__str__(), H_gate_inverse.__str__())

        # Create a parameterized rotation on qubit 1 with angle 2 radians
        RX_gate = Gate("RX", 1, parameter=2.)
        RX_gate_inverse = Gate("RX", 1, parameter=-2.)
        self.assertEqual(RX_gate.inverse().__str__(), RX_gate_inverse.__str__())

        # Create a parameterized rotation on qubit 1 , with an undefined angle, that will be variational
        RZ_gate = Gate("RZ", 1, parameter="an expression", is_variational=True)
        with self.assertRaises(AttributeError):
            RZ_gate.inverse()

    def test_incorrect_gate(self):
        """ Test to catch a gate with inputs that do not make sense """

        self.assertRaises(ValueError, Gate, "H", -1)
        self.assertRaises(ValueError, Gate, "CNOT", 0, control=0.3)
        self.assertRaises(ValueError, Gate, 'X', target=0, control=1)
        self.assertRaises(ValueError, Gate, "CNOT", target=0, control=0)

    def test_integer_types(self):
        """ Test to catch error with incorrect target or control qubit index type"""
        self.assertRaises(ValueError, Gate, "CSWAP", target=[0, 'a'], control=np.array([1], dtype=np.int32))
        self.assertRaises(ValueError, Gate, "X", target=0, control=[-1, 2, 3],)

    def test_gate_equality(self):
        """ Test behaviour of == and != operators on gates """
        g1 = Gate("CPOTATO", target=2, control=0, parameter=0, is_variational=True)
        g2 = Gate("CPOTATO", target=2, control=0, parameter="", is_variational=True)
        g3 = Gate("CPOTATO", target=2, control=0, parameter=0, is_variational=True)

        self.assertTrue(g1 == g3)
        self.assertTrue(g1 != g2)

    def test_gate_equality_modulo_twopi(self):
        """ Test behaviour of == with parameter outside [0, 2pi]. """
        g1 = Gate("POTATO", target=0, parameter=np.pi, is_variational=True)
        g2 = Gate("POTATO", target=0, parameter=3*np.pi, is_variational=True)
        g3 = Gate("POTATO", target=0, parameter=-np.pi, is_variational=True)

        self.assertTrue(g1 == g2)
        self.assertTrue(g1 == g3)
        self.assertTrue(g2 == g3)

    def test_too_many_qubits_on_gates(self):
        """ Test the behavior when too many qubits are selected for a gate. """

        # Try to create a Hadamard gate acting on qubits 0 and 1.
        self.assertRaises(ValueError, Gate, "H", target=[0, 1])

        # Try to create a XX gate acting on qubits 0, 1 and 2.
        self.assertRaises(ValueError, Gate, "XX", target=[0, 1, 2])


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from tangelo.helpers.utils import is_package_installed
from tangelo.molecule_library import mol_pyridine


class SemiEmpiricalSolverTest(unittest.TestCase):

    @unittest.skipIf(not is_package_installed("pyscf.semiempirical"), "Test Skipped: pyscf.semiempirical module not available \n")
    def test_mindo3_energy(self):
        """Test MINDO3Solver with pyridine. Validated with:
            - MINDO/3-derived geometries and energies of alkylpyridines and the
            related N-methylpyridinium cations. Jeffrey I. Seeman,
            John C. Schug, and Jimmy W. Viers
            The Journal of Organic Chemistry 1983 48 (14), 2399-2407
            DOI: 10.1021/jo00162a021.
        """
        from tangelo.algorithms.classical import MINDO3Solver

        solver = MINDO3Solver(mol_pyridine)
        energy = solver.simulate()

        self.assertAlmostEqual(energy, -33.04112644117467, places=6)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines the qubit coupled cluster ansatz class with involutory
linear combinations (ILC) of anticommuting sets (ACS) of Pauli words
(generators). Relative to the direct interation set (DIS) of QCC generators,
which incur an exponential growth of Hamiltonian terms upon dressing, the ACS
of ILC generators enables Hamiltonian dressing such that the number of terms
grows quadratically and exact quadratic truncation of the Baker-Campbell-Hausdorff
expansion. For more information about this ansatz, see references below.

Refs:
    1. R. A. Lang, I. G. Ryabinkin, and A. F. Izmaylov.
        arXiv:2002.05701v1, 2020, 1–10.
    2. R. A. Lang, I. G. Ryabinkin, and A. F. Izmaylov.
        J. Chem. Theory Comput. 2021, 17, 1, 66–78.
"""

import warnings

import numpy as np

from tangelo.toolboxes.qubit_mappings.mapping_transform import get_qubit_number,\
                                                               fermion_to_qubit_mapping
from tangelo.linq import Circuit
from tangelo.toolboxes.ansatz_generator.ansatz import Ansatz
from tangelo.toolboxes.ansatz_generator.ansatz_utils import exp_pauliword_to_gates
from tangelo.toolboxes.ansatz_generator._qubit_mf import init_qmf_from_hf, get_qmf_circuit, purify_qmf_state
from tangelo.toolboxes.ansatz_generator._qubit_cc import construct_dis
from tangelo.toolboxes.ansatz_generator._qubit_ilc import construct_acs, get_ilc_params_by_diag


class ILC(Ansatz):
    """This class implements the ILC ansatz. Closed-shell and restricted open-shell ILC are
    supported. While the form of the ILC ansatz is the same for either variation, the underlying
    fermionic mean-field state is treated differently depending on the spin. Closed-shell
    or restricted open-shell ILC implies that spin = 0 or spin != 0 and the fermionic mean-field
    state is obtained using a RHF or ROHF Hamiltonian, respectively.

    Args:
        molecule (SecondQuantizedMolecule): The molecular system.
        mapping (str): One of the supported  mapping identifiers. Default, "jw".
        up_then_down (bool): Change basis ordering putting all spin-up orbitals first,
            followed by all spin-down. Default, False.
        acs (list of QubitOperator): The mutually anticommuting generator list for the ILC ansatz.
            Default, None.
        qmf_circuit (Circuit): An instance of tangelo.linq Circuit class implementing a QMF state
            circuit. If passed from the QMF ansatz class, parameters are variational.
            If None, one is created with QMF parameters that are not variational. Default, None.
        qmf_var_params (list or numpy array of float): QMF variational parameter set.
            If None, the values are determined using a Hartree-Fock reference state. Default, None.
        qubit_ham (QubitOperator): Pass a qubit Hamiltonian to the  ansatz class and ignore
            the fermionic Hamiltonian in molecule. Default, None.
        deilc_dtau_thresh (float): Threshold for |dEILC/dtau| so that a candidate group is added
            to the DIS if |dEILC/dtau| >= deilc_dtau_thresh for a generator. Default, 1e-3 a.u.
        ilc_tau_guess (float): The initial guess for all ILC variational parameters.
            Default, 1e-2 a.u.
        max_ilc_gens (int or None): Maximum number of generators allowed in the ansatz. If None,
            one generator from each DIS group is selected. If int, then min(|DIS|, max_ilc_gens)
            generators are selected in order of decreasing |dEILC/dtau|. Default, None.
    """

    def __init__(self, molecule, mapping="jw", up_then_down=False, acs=None,
                 qmf_circuit=None, qmf_var_params=None, qubit_ham=None, ilc_tau_guess=1e-2,
                 deilc_dtau_thresh=1e-3, max_ilc_gens=None):

        if not molecule:
            raise ValueError("An instance of SecondQuantizedMolecule is required for initializing "
                             "the self.__class__.__name__ ansatz class.")
        self.molecule = molecule
        self.mapping = mapping
        self.up_then_down = up_then_down
        if self.mapping.lower() == "jw" and not self.up_then_down:
            warnings.warn("Spin-orbital ordering shifted to all spin-up first then down to "
                          "ensure efficient generator screening for the Jordan-Wigner mapping "
                          "with the self.__class__.__name__ ansatz.", RuntimeWarning)
            self.up_then_down = True

        self.n_spinorbitals = self.molecule.n_active_sos
        if self.n_spinorbitals % 2 != 0:
            raise ValueError("The total number of spin-orbitals should be even.")

        self.spin = molecule.spin
        self.fermi_ham = self.molecule.fermionic_hamiltonian
        self.n_electrons = self.molecule.n_electrons
        self.n_qubits = get_qubit_number(self.mapping, self.n_spinorbitals)

        self.qubit_ham = qubit_ham
        if qubit_ham is None:
            self.fermi_ham = self.molecule.fermionic_hamiltonian
            self.qubit_ham = fermion_to_qubit_mapping(self.fermi_ham, self.mapping,
                                                      self.n_spinorbitals, self.n_electrons,
                                                      self.up_then_down, self.spin)

        self.qmf_var_params = qmf_var_params
        if self.qmf_var_params is None:
            self.qmf_var_params = init_qmf_from_hf(self.n_spinorbitals, self.n_electrons,
                                                   self.mapping, self.up_then_down, self.spin)
        elif isinstance(self.qmf_var_params, list):
            self.qmf_var_params = np.array(self.qmf_var_params)
        if self.qmf_var_params.size != 2 * self.n_qubits:
            raise ValueError("The number of QMF variational parameters must be 2 * n_qubits.")

        self.qmf_circuit = qmf_circuit

        self.acs = acs
        self.ilc_tau_guess = ilc_tau_guess
        self.deilc_dtau_thresh = deilc_dtau_thresh
        self.max_ilc_gens = max_ilc_gens

        # Get purified QMF parameters and build the DIS & ACS or use a list of generators.
        if self.acs is None:
            pure_var_params = purify_qmf_state(self.qmf_var_params, self.n_spinorbitals,
                                               self.n_electrons, self.mapping, self.up_then_down, self.spin)
            self.dis = construct_dis(self.qubit_ham, pure_var_params, self.deilc_dtau_thresh)
            self.max_ilc_gens = len(self.dis) if self.max_ilc_gens is None\
                                else min(len(self.dis), self.max_ilc_gens)
            self.acs = construct_acs(self.dis, self.max_ilc_gens, self.n_qubits)
            self.n_var_params = len(self.acs)
        else:
            self.dis = None
            self.n_var_params = len(self.acs)

        # Supported reference state initialization
        self.supported_reference_state = {"HF"}
        # Supported var param initialization
        self.supported_initial_var_params = {"qmf_state", "ilc_tau_guess", "random", "diag"}

        # Default starting parameters for initialization
        self.default_reference_state = "HF"
        self.var_params_default = "diag"
        self.var_params = None
        self.rebuild_dis = False
        self.rebuild_acs = False
        self.ilc_circuit = None
        self.circuit = None

    def set_var_params(self, var_params=None):
        """Set values for variational parameters, such as zeros or floats,
        providing some keywords for users, and also supporting direct user input
        (list or numpy array). Return the parameters so that workflows such as VQE can
        retrieve these values. """

        if var_params is None:
            var_params = self.var_params_default

        if isinstance(var_params, str):
            var_params = var_params.lower()
            if var_params not in self.supported_initial_var_params:
                raise ValueError(f"Supported keywords for initializing variational parameters: "
                                 f"{self.supported_initial_var_params}")
            # Initialize the ILC wave function as |ILC> = |QMF>
            if var_params == "qmf_state":
                initial_var_params = np.zeros((self.n_var_params,), dtype=float)
            # Initialize all ILC parameters to the same value specified by self.ilc_tau_guess
            elif var_params == "ilc_tau_guess":
                initial_var_params = self.ilc_tau_guess * np.ones((self.n_var_params,))
            # Initialize tau parameters randomly over the domain [0., 2 pi)
            elif var_params == "random":
                initial_var_params = 2. * np.pi * np.random.random((self.n_var_params,))
            # Initialize ILC parameters by matrix diagonalization (see Appendix B, Refs. 1 & 2).
            elif var_params == "diag":
                initial_var_params = get_ilc_params_by_diag(self.qubit_ham, self.acs, self.qmf_var_params)
        else:
            initial_var_params = np.array(var_params)
            if initial_var_params.size != self.n_var_params:
                raise ValueError(f"Expected {self.n_var_params} variational parameters but "
                                 f"received {initial_var_params.size}.")
        self.var_params = initial_var_params
        return initial_var_params

    def prepare_reference_state(self):
        """Returns circuit preparing the reference state of the ansatz (e.g prepare reference
        wavefunction with HF, multi-reference state, etc). These preparations must be consistent
        with the transform used to obtain the  operator. """

        if self.default_reference_state not in self.supported_reference_state:
            raise ValueError(f"Only supported reference state methods are: "
                             f"{self.supported_reference_state}.")
        if self.default_reference_state == "HF":
            reference_state_circuit = get_qmf_circuit(self.qmf_var_params, False)
        return reference_state_circuit

    def build_circuit(self, var_params=None):
        """Build and return the quantum circuit implementing the state preparation ansatz
         (with currently specified initial_state and var_params). """

        if var_params is not None:
            self.set_var_params(var_params)
        elif self.var_params is None:
            self.set_var_params()

        # Build a QMF state preparation circuit
        if self.qmf_circuit is None:
            self.qmf_circuit = self.prepare_reference_state()

        # Build create the list of ILC qubit operators
        self.ilc_op_list = self._get_ilc_op()

        # Obtain quantum circuit through trotterization of the list of ILC operators
        pauli_word_gates = []
        for ilc_op in self.ilc_op_list:
            pauli_word, coef = list(ilc_op.terms.items())[0]
            pauli_word_gates += exp_pauliword_to_gates(pauli_word, coef, variational=True)
        self.ilc_circuit = Circuit(pauli_word_gates)
        self.circuit = self.qmf_circuit + self.ilc_circuit if self.qmf_circuit.size != 0\
                       else self.ilc_circuit

    def update_var_params(self, var_params):
        """Shortcut: set value of variational parameters in the existing ansatz circuit member.
        Preferable to rebuilding your circuit from scratch, which can be an involved process.
        """

        # Update the ILC variational parameters
        self.set_var_params(var_params)

        # Build the ILC ansatz operator
        self.ilc_op_list = self._get_ilc_op()

        pauli_word_gates = []
        for ilc_op in self.ilc_op_list:
            pauli_word, coef = list(ilc_op.terms.items())[0]
            pauli_word_gates += exp_pauliword_to_gates(pauli_word, coef, variational=True)
        self.ilc_circuit = Circuit(pauli_word_gates)
        self.circuit = self.qmf_circuit + self.ilc_circuit if self.qmf_circuit.size != 0\
                       else self.ilc_circuit

    def _get_ilc_op(self):
        """Returns the ILC operators ordered according to the argument of
            Eq. C1, Appendix C, Ref. 1.

        Returns:
            list of QubitOperator: the list of ILC qubit operators
        """

        # Rebuild DIS & ACS in case qubit_ham changed or they and qubit_op_list don't exist
        if self.rebuild_dis or self.rebuild_acs or not self.acs:
            pure_var_params = purify_qmf_state(self.qmf_var_params, self.n_spinorbitals,
                                               self.n_electrons, self.mapping, self.up_then_down, self.spin)
            self.dis = construct_dis(self.qubit_ham, pure_var_params, self.deilc_dtau_thresh)
            self.max_ilc_gens = len(self.dis) if self.max_ilc_gens is None\
                                else min(len(self.dis), self.max_ilc_gens)
            self.acs = construct_acs(self.dis, self.max_ilc_gens, self.n_qubits)
            self.ilc_op_list = None

        # Build the ILC qubit operator list
        ilc_op_list = []
        for i in range(self.n_var_params - 1, 0, -1):
            ilc_op_list.append(-0.5 * self.var_params[i] * self.acs[i])
        ilc_op_list.append(-self.var_params[0] * self.acs[0])
        for i in range(1, self.n_var_params):
            ilc_op_list.append(-0.5 * self.var_params[i] * self.acs[i])
        return ilc_op_list



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import numpy as np

from tangelo.molecule_library import mol_H4_doublecation_minao, mol_H4_doublecation_321g, mol_H10_321g, mol_H10_minao
from tangelo.problem_decomposition import DMETProblemDecomposition
from tangelo.problem_decomposition.dmet import Localization
from tangelo.algorithms.variational import VQESolver
from tangelo.toolboxes.molecular_computation.rdms import matricize_2rdm


class DMETProblemDecompositionTest(unittest.TestCase):

    def test_incorrect_number_atoms(self):
        """Tests if the program raises the error when the number of fragment
        sites is not equal to the number of atoms in the molecule.
        """

        opt_dmet = {"molecule": mol_H10_321g,
                    "fragment_atoms": [1, 1, 1, 1],
                    "electron_localization": Localization.meta_lowdin,
                    "verbose": False
                    }

        # The molecule has more atoms than this.
        with self.assertRaises(RuntimeError):
            DMETProblemDecomposition(opt_dmet)

    def test_incorrect_number_solvers(self):
        """Tests if the program raises the error when the number of fragment
        sites is not equal to the number of solvers.
        """

        opt_dmet = {"molecule": mol_H10_321g,
                    "fragment_atoms": [2, 3, 2, 3],
                    "fragment_solvers": ["fci", "fci"],
                    "verbose": False
                    }

        with self.assertRaises(RuntimeError):
            DMETProblemDecomposition(opt_dmet)

    def test_h10ring_ml_fci(self):
        """ Tests the result from DMET against a value from a reference
        implementation with meta-lowdin localization and FCI solution to
        fragments."""

        opt_dmet = {"molecule": mol_H10_321g,
                    "fragment_atoms": [1]*10,
                    "fragment_solvers": "fci",
                    "electron_localization": Localization.meta_lowdin,
                    "verbose": False
                    }

        dmet_solver = DMETProblemDecomposition(opt_dmet)
        dmet_solver.build()
        energy = dmet_solver.simulate()

        self.assertAlmostEqual(energy, -4.498973024, places=4)

    def test_h4ring_ml_ccsd_minao(self):
        """Tests the result from DMET against a value from a reference
        implementation with meta-lowdin localization and CCSD solution to
        fragments.
        """

        opt_dmet = {"molecule": mol_H4_doublecation_minao,
                    "fragment_atoms": [1, 1, 1, 1],
                    "fragment_solvers": "ccsd",
                    "electron_localization": Localization.meta_lowdin,
                    "verbose": False
                    }

        dmet_solver = DMETProblemDecomposition(opt_dmet)
        dmet_solver.build()
        energy = dmet_solver.simulate()

        self.assertAlmostEqual(energy, -0.854379, places=6)

    def test_h4ring_ml_default_minao(self):
        """Tests the result from DMET against a value from a reference
        implementation with meta-lowdin localization and default solver
        (currently CCSD) for fragments.
        """

        opt_dmet = {"molecule": mol_H4_doublecation_minao,
                    "fragment_atoms": [1, 1, 1, 1],
                    "electron_localization": Localization.meta_lowdin,
                    "verbose": False
                    }

        dmet_solver = DMETProblemDecomposition(opt_dmet)
        dmet_solver.build()
        energy = dmet_solver.simulate()

        self.assertAlmostEqual(energy, -0.854379, places=6)

    def test_h4ring_ml_fci_minao(self):
        """ Tests the result from DMET against a value from a reference
        implementation with meta-lowdin localization and FCI solution to
        fragments.
        """

        opt_dmet = {"molecule": mol_H4_doublecation_minao,
                    "fragment_atoms": [1, 1, 1, 1],
                    "fragment_solvers": "fci",
                    "electron_localization": Localization.meta_lowdin,
                    "verbose": False
                    }

        dmet_solver = DMETProblemDecomposition(opt_dmet)
        dmet_solver.build()
        energy = dmet_solver.simulate()

        self.assertAlmostEqual(energy, -0.854379, places=4)

    def test_solver_mix(self):
        """Tests that solving with multiple solvers works.

        With this simple system, we can assume that both CCSD and FCI can reach
        chemical accuracy.
        """

        opt_dmet = {"molecule": mol_H4_doublecation_321g,
                    "fragment_atoms": [1, 1, 1, 1],
                    "fragment_solvers": ["fci", "fci", "ccsd", "ccsd"],
                    "electron_localization": Localization.iao,
                    "verbose": False
                    }

        solver = DMETProblemDecomposition(opt_dmet)
        solver.build()
        energy = solver.simulate()
        self.assertAlmostEqual(energy, -0.94199, places=4)

    def test_fragment_ids(self):
        """Tests if a nested list of atom ids is provided."""

        opt_dmet = {"molecule": mol_H4_doublecation_321g,
                    "fragment_atoms": [[0], [1], [2], [3]],
                    "fragment_solvers": "ccsd",
                    "electron_localization": Localization.iao,
                    "verbose": False
                    }

        solver = DMETProblemDecomposition(opt_dmet)

        self.assertEqual(solver.fragment_atoms, [1, 1, 1, 1])

    def test_build_with_atom_indices(self):
        """Tests if a mean field is recomputed when providing atom indices."""

        opt_dmet = {"molecule": mol_H4_doublecation_321g,
                    "fragment_atoms": [[0], [1], [2], [3]],
                    "fragment_solvers": "ccsd",
                    "electron_localization": Localization.iao,
                    "verbose": False
                    }

        solver = DMETProblemDecomposition(opt_dmet)
        solver.build()

    def test_fragment_ids_exceptions(self):
        """Tests exceptions if a bad nested list of atom ids is provided. Two
        cases: if an atom id is higher than the number of atoms and if an id is
        detected twice (or more).
        """

        opt_dmet = {"molecule": mol_H4_doublecation_321g,
                    "fragment_atoms": [[0, 0], [1], [2], [3]],
                    "fragment_solvers": "ccsd",
                    "electron_localization": Localization.iao,
                    "verbose": False
                    }

        with self.assertRaises(RuntimeError):
            DMETProblemDecomposition(opt_dmet)

        opt_dmet["fragment_atoms"] = [[0], [1], [2], [4]]

        with self.assertRaises(RuntimeError):
            DMETProblemDecomposition(opt_dmet)

    def test_retrieving_quantum_data(self):
        """Test if getting back a fragment gives the same RDMs."""

        opt_dmet = {"molecule": mol_H10_minao,
                    "fragment_atoms": [1]*10,
                    "fragment_solvers": ["vqe"] + ["ccsd"]*9,
                    "electron_localization": Localization.meta_lowdin,
                    "verbose": True,
                    "solvers_options": [{"qubit_mapping": "scBK",
                                         "initial_var_params": "ones",
                                         "up_then_down": True,
                                         "verbose": False}] + [{}]*9,
                    }

        dmet_solver = DMETProblemDecomposition(opt_dmet)
        dmet_solver.build()

        # One shot loop with the optimal chemical potential.
        dmet_solver.n_iter = 0
        dmet_solver._oneshot_loop(-0.0000105903, save_results=True)

        ref_onerdm = [[1.99836090e+00, 6.91189485e-04],
                      [6.91189485e-04, 1.63910482e-03]]
        ref_twordm = [[[[ 1.99836064e+00,  7.11566876e-04],
                        [ 7.11566876e-04,  2.53371393e-07]],
                        [[ 7.11566876e-04, -5.72277575e-02],
                        [ 2.53371393e-07, -2.03773912e-05]]],
                        [[[ 7.11566876e-04,  2.53371393e-07],
                        [-5.72277575e-02, -2.03773912e-05]],
                        [[ 2.53371393e-07, -2.03773912e-05],
                        [-2.03773912e-05,  1.63885145e-03]]]]

        fragment, _, q_circuit = dmet_solver.quantum_fragments_data[0]

        vqe_solver = VQESolver({"molecule": fragment, "ansatz": q_circuit,
                                "qubit_mapping": "scBK"})
        vqe_solver.build()
        vqe_solver.simulate()

        onerdm, twordm = vqe_solver.get_rdm(vqe_solver.optimal_var_params)

        # Test traces of matrices
        n_elec = fragment.n_active_electrons
        n_orb = fragment.n_active_sos // 2
        self.assertAlmostEqual(np.trace(onerdm), n_elec, delta=1e-3,
                               msg="Trace of one_rdm does not match number of electrons")

        rho = matricize_2rdm(twordm, n_orb)
        self.assertAlmostEqual(np.trace(rho), n_elec * (n_elec - 1), delta=1e-3,
                               msg="Trace of two_rdm does not match n_elec * (n_elec-1)")


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions to generate FermionOperators corresponding to the
generalized UCCSD (UCCGSD) ansatz for a designated number of orbitals or qubits.
Excitations are grouped together to minimize redundancy and symmetry-violating
independence between related spin-orbitals. Groups are specified here as defined
in Tang et al (arXiv:1911.10205).
"""

import numpy as np
import itertools

from tangelo.toolboxes.operators import FermionOperator


def hermitian_conjugate(terms):
    r"""Create nested list of tuples which can be utilized to generate
    FermionOperator associated with the Hermitian conjugate of those same
    operators implied by input *terms*.

    For a product of fermionic terms, the Hermitian conjugate corresponds to a
    reverse ordering of the operators. The prefactor is the same, up to an
    overall sign, dictated by the number of permutations.

    For singles and doubles, permutations are odd, so sign of prefactor is
    flipped.

    Args:
        terms (list): list of lists. Each sub-list is formatted as a tuple, and
            a tuple and a float. The tuples are themselves, nested tuple
            dictating spin-orbital index (int), and creation (1) or annihilation
            (0). e.g. one element in the terms list may be:
            [((0,1),(1,0),(3,1),(4,0)),0.5] =
            0.5a_0^{\dagger}a_1a_3^{\dagger}a_4

    Returns:
        list: same format as input terms, with orbitals and operations inverted,
            and sign flipped.
    """
    try:
        return [[((v[0][3][0], 1),
                  (v[0][2][0], 1),
                  (v[0][1][0], 0),
                  (v[0][0][0], 0)),
                 -v[1]] for v in terms]
    except Exception:
        raise ValueError("Input terms must be format as, e.g. [[((int,1),(int,0),(int,1),(int,0)),float],...]")


def get_spin_ordered(n_orbs, pp, qq, rr=-1, ss=-1, up_down=False):
    """For a given set of orbitals, provide corresponding spin-orbital indices,
    based on the desired spin-orbital ordering. Depending on user-convention,
    one typically orders spin-orbitals as all spin-up followed by all spin-down
    (this is the convention followed in e.g. qiskit), or alternating spin-up,
    spin-down (as done in e.g. openfermion). By setting *up_down* flag to True,
    one uses the qiskit convention, and False, the openfermion convention.

    This function accommodates either a pair or quartet of orbital indices, and
    returns two tuples of int reflecting the spin-up fermion indices for desired
    orbitals, and the spin-down fermion indices for the same.

    Args:
        n_orbs (int): number of orbitals in the fermion basis (this is number of
            spin-orbitals divided by 2).
        pp (int): orbital index.
        qq (int): orbital index.
        rr (int): orbital index, only specified for four-fermion operator.
        ss (int): orbital index, only specified for four-fermion operator.
        up_down (bool): flag for using qiskit (True) or openfermion (False)
            spin-ordering conventions.

    Returns:
        up (tuple): tuple of int -- 4 if rr and ss kwargs specified, else 2
            corresponding to spin-up spin-orbitals.
        down (tuple): tuple of int -- 4 if rr and ss kwargs specified, else 2
            corresponding to spin-down spin-orbitals.
    """
    if type(n_orbs) != int:
        raise TypeError("Invalid datatype for number of orbitals.")
    if n_orbs < 1:
        raise ValueError("Number of orbitals must be positive.")
    try:
        pp, qq, rr, ss = int(pp), int(qq), int(rr), int(ss)  # Force orbital indices to int-type
    except TypeError:
        raise TypeError("All orbital indices (pp, qq, rr, ss) must be integer-type.")

    if type(up_down) != bool:
        raise TypeError("Spin-ordering arg (up_down) must be boolean.")

    # all spin up then all spin down
    if up_down:
        up = pp, qq, rr, ss
        down = n_orbs + pp, n_orbs + qq, n_orbs + rr, n_orbs + ss
    # alternating spin up/spin down
    else:
        up = 2 * pp, 2 * qq, 2 * rr, 2 * ss
        down = 2 * pp + 1, 2 * qq + 1, 2 * rr + 1, 2 * ss + 1
    # if user has passed a 2-fermion operator
    if rr < 0:
        return up[:2], down[:2]
    # otherwise, a user has passed a 4-fermion operator
    return up, down


def get_group_1_2(n_orbs, p, q, r, s, up_down=False):
    r"""Identify spin singlet and spin triplet -type 4-orbital excitations. We
    group equivalent terms,along with their Hermitian conjugates to prepare a
    single FermionOperator which will obey spin-symmetries. We then generalize
    expressions of the form:

    a_p^{\dagger}a_q^{\dagger}a_r a_s with the constraints

    Either: (p < q and p < r and p < s and r < s) and (q != r and q != s)
    Or: (q == r and p < s) and (pp != qq and ss != qq).

    Args:
        n_orbs (int): number of orbitals in basis (this is number of
            spin-orbitals divided by 2).
        p (int): orbital index.
        q (int): orbital index.
        r (int): orbital index.
        s (int): orbital index.
        up_down (bool): flag for using qiskit (True) or openfermion (False)
            spin-ordering conventions.

    Returns:
        (list): list of input for FermionOperator corresponding to all singlet
            and triplet terms.
    """
    if type(n_orbs) != int:
        raise TypeError("Number of orbitals (n_orbs) must be integer-type.")
    if n_orbs < 1:
        raise ValueError("Number of orbitals (n_orbs) must be at least 1.")
    if type(up_down) != bool:
        raise TypeError("Spin-ordering arg (up_down) must be boolean.")

    up, dn = get_spin_ordered(n_orbs, p, q, r, s, up_down=up_down)  # get spin-orbital indices

    # get triplet term
    triplet = [[((up[0], 1), (up[1], 1), (up[2], 0), (up[3], 0)), 1.0],
               [((up[0], 1), (dn[1], 1), (up[2], 0), (dn[3], 0)), 0.5],
               [((up[0], 1), (dn[1], 1), (dn[2], 0), (up[3], 0)), 0.5],
               [((dn[0], 1), (up[1], 1), (up[2], 0), (dn[3], 0)), 0.5],
               [((dn[0], 1), (up[1], 1), (dn[2], 0), (up[3], 0)), 0.5],
               [((dn[0], 1), (dn[1], 1), (dn[2], 0), (dn[3], 0)), 1.0]]

    # get singlet term
    singlet = [[((up[0], 1), (dn[1], 1), (up[2], 0), (dn[3], 0)), 0.5],
               [((up[0], 1), (dn[1], 1), (dn[2], 0), (up[3], 0)), -0.5],
               [((dn[0], 1), (up[1], 1), (up[2], 0), (dn[3], 0)), -0.5],
               [((dn[0], 1), (up[1], 1), (dn[2], 0), (up[3], 0)), 0.5]]

    # add Hermitian conjugates
    triplet += hermitian_conjugate(triplet)
    singlet += hermitian_conjugate(singlet)

    return singlet + triplet


def get_group_3_4(n_orbs, p, q, r, s, up_down=False):
    r"""Identify spin singlet -type 4-orbital excitations. We group equivalent
    terms, along with their Hermitian conjugates to prepare a single
    FermionOperator which will obey spin-symmetries. We then generalize
    expressions of the form:

    a_p^{\dagger}a_q^{\dagger}a_r a_s with the constraints

    Either: (p == q and q != r and r != s and s != p) and r < s
    Or: p == q and q == r and r != s.

    Note there is no spin-triplet term for these criteria.

    Args:
        n_orbs (int): number of orbitals in basis (this is number of
            spin-orbitals divided by 2).
        p (int): orbital index.
        q (int): orbital index.
        r (int): orbital index.
        s (int): orbital index.
        up_down (bool): flag for using qiskit (True) or openfermion (False)
            spin-ordering conventions.

    Returns:
        singlet (list): list of input for FermionOperator corresponding to
            spin-singlet terms.
    """
    if type(n_orbs) != int:
        raise TypeError("Number of orbitals (n_orbs) must be integer-type.")
    if n_orbs < 1:
        raise ValueError("Number of orbitals (n_orbs) must be at least 1.")
    if type(up_down) != bool:
        raise TypeError("Spin-ordering arg (up_down) must be boolean.")

    up, dn = get_spin_ordered(n_orbs, p, q, r, s, up_down=up_down)  # get spin-orbital indices

    # prepare combinations
    singlet = [[((up[0], 1), (dn[1], 1), (up[2], 0), (dn[3], 0)), 1.0],
               [((up[0], 1), (dn[1], 1), (dn[2], 0), (up[3], 0)), 1.0]]

    # add Hermitian conjugate
    singlet += hermitian_conjugate(singlet)

    return singlet


def get_group_5(n_orbs, p, q, r, s, up_down=False):
    r"""Identify spin singlet -type 4-orbital excitations. We group equivalent
    terms, along with their Hermitian conjugates to prepare a single
    FermionOperator which will obey spin-symmetries. We then generalize
    expressions of the form:

    a_p^{\dagger}a_q^{\dagger}a_r a_s with the constraints

    p == q and q != r and r == s and p < s

    Note there is no spin-triplet term for this criteria.

    Args:
        n_orbs (int): number of orbitals in basis (this is number of
            spin-orbitals divided by 2).
        p (int): orbital index.
        q (int): orbital index.
        r (int): orbital index.
        s (int): orbital index.
        up_down (bool): flag for using qiskit (True) or openfermion (False)
            spin-ordering conventions.

    Returns:
        singlet (list): list of input for FermionOperator corresponding to
            spin-singlet terms.
    """
    if type(n_orbs) != int:
        raise TypeError("Number of orbitals (n_orbs) must be integer-type.")
    if n_orbs < 1:
        raise ValueError("Number of orbitals (n_orbs) must be at least 1.")
    if type(up_down) != bool:
        raise TypeError("Spin-ordering arg (up_down) must be boolean.")

    # get spin-orbital indices
    up, dn = get_spin_ordered(n_orbs, p, q, r, s, up_down=up_down)

    # prepare combinations
    singlet = [[((up[0], 1), (dn[1], 1), (up[2], 0), (dn[3], 0)), 2.0]]

    # add Hermitian conjugate terms
    singlet += hermitian_conjugate(singlet)

    return singlet


def get_doubles(n_orbs, up_down=False):
    r"""Prepare a list of all double-excitation terms in the UCCGSD for a given
    number of fermionic orbital basis states, grouped together according to
    spin-compensated spin-triplet or spin-singlet spin-orbital combinations. The
    number of combinations of spin-orbitals for a given combination of orbital
    basis states, as well as the relative prefactors, depend on the specific
    orbital indices. For example, there are more ways to permute four distinct
    orbitals, than when two of the orbital indices are the same.

    Five different groupings of orbital indices can be identified, as in e.g.
    (arXiv:1911.10205) covering all possible combinations.

    Args:
        n_orbs (int): number of orbitals in basis (this is number of
            spin-orbitals divided by 2).
        up_down (bool): flag for using qiskit (True) or openfermion (False)
            spin-ordering conventions.

    Returns:
        all_terms (list): list of lists. Each sub-list is formatted as a tuple,
            and a tuple and a float. The tuples are themselves, nested tuple
            dictating spin-orbital index (int), and creation (1) or annihilation
            (0). e.g. one element in the all_terms list may be:
            [((0,1),(1,0),(3,1),(4,0)),0.5] =
            0.5a_0^{\dagger}a_1a_3^{\dagger}a_4
    """
    if type(n_orbs) != int:
        raise TypeError("Number of orbitals (n_orbs) must be integer-type.")
    if n_orbs < 1:
        raise ValueError("Number of orbitals (n_orbs) must be at least 1.")
    if type(up_down) != bool:
        raise TypeError("Spin-ordering arg (up_down) must be boolean.")

    all_terms = list()
    selection = np.linspace(0, n_orbs - 1, n_orbs, dtype=int)
    for pp, qq, rr, ss in itertools.product(selection, repeat=4):

        # generate terms according to relationship between the spin-orbitals involved.
        if (pp < qq and pp < rr and pp < ss) and (rr < ss) and (qq != rr and qq != ss):
            terms = get_group_1_2(n_orbs, pp, qq, rr, ss, up_down=up_down)
        elif qq == rr and pp < ss and pp != qq and ss != qq:
            terms = get_group_1_2(n_orbs, pp, qq, rr, ss, up_down=up_down)
        elif (pp == qq and qq != rr and rr != ss and ss != pp) and rr < ss:
            terms = get_group_3_4(n_orbs, pp, qq, rr, ss, up_down=up_down)
        elif pp == qq and qq == rr and rr != ss:
            terms = get_group_3_4(n_orbs, pp, qq, rr, ss, up_down=up_down)
        elif pp == qq and qq != rr and rr == ss and pp < ss:
            terms = get_group_5(n_orbs, pp, qq, rr, ss, up_down=up_down)
        else:
            # skip the append if no new terms defined by conditions above
            continue

        all_terms.append(terms)

    return all_terms


def get_singles(n_orbs, up_down=False):
    r"""Prepare a list of all single-excitation terms in the UCCGSD for a given
    number of fermionic orbital basis states, grouped together according to
    spin-compensated spin-orbital combinations.

    Args:
        n_orbs (int): number of orbitals in basis (this is number of
            spin-orbitals divided by 2).
        up_down (bool): flag for using qiskit (True) or openfermion (False)
            spin-ordering conventions.

    Returns:
        all_terms (list): list of lists. Each sub-list is formatted as a tuple,
            and a tuple and a float. The tuples are themselves, nested tuple
            dictating spin-orbital index (int), and creation (1) or annihilation
            (0). e.g. one element in the all_terms list may be:
            [((0,1),(1,0)),0.5] = 0.5a_0^{\dagger}a_1
    """
    if type(n_orbs) != int:
        raise TypeError("Number of orbitals (n_orbs) must be integer-type.")
    if n_orbs < 1:
        raise ValueError("Number of orbitals (n_orbs) must be at least 1.")
    if type(up_down) != bool:
        raise TypeError("Spin-ordering arg (up_down) must be boolean.")

    all_terms = list()
    selection = np.linspace(0, n_orbs - 1, n_orbs, dtype=int)  # all possible orbitals
    for pp, qq in itertools.product(selection, repeat=2):  # iterate over all pairings of orbital-indices

        # avoid duplicates, take only lower-triangle
        if qq <= pp:
            continue

        # get spin-orbital indices
        up, down = get_spin_ordered(n_orbs, pp, qq, up_down=up_down)

        terms = [[((up[0], 1), (up[1], 0)), 1.],
                 [((down[0], 1), (down[1], 0)), 1.],  # spin-compensated (up <> down)
                 [((up[1], 1), (up[0], 0)), -1.],  # Hermitian conjugate
                 [((down[1], 1), (down[0], 0)), -1.]]  # h.c. of spin-compensated term

        all_terms.append(terms)

    return all_terms


def get_all_excitations(n_orbs, up_down=False):
    """Enumerate all possible single and double excitations in the UCCGSD for a
    given number of fermionic orbital basis states. Prepare list of inputs to be
    applied as FermionicOperators.

    The output is a third-order nested list. The list represents groups of
    operators acting on related sets of orbitals. Each of these groups
    enumerate a single or double excitations, and the related coefficients
    between their constituent terms. This output is used to build a set of
    FermionOperator objects.

    Args:
        n_orbs (int): number of orbitals in basis (this is number of
            spin-orbitals divided by 2).
        up_down (bool): flag for using qiskit (True) or openfermion (False)
            spin-ordering conventions.

    Returns:
        all_terms (list): nested list as detailed above. Each element is a list
            over either a single or double excitation, represented by
            coefficients and the related operator indices and actions
            (creation/annihilation).
    """
    if type(n_orbs) != int:
        raise TypeError("Number of orbitals (n_orbs) must be integer type.")
    if n_orbs < 1:
        raise ValueError("Number of orbitals (n_orbs) must be at least 1.")
    if type(up_down) != bool:
        raise TypeError("Spin-ordering arg (up_down) must be boolean.")

    singles = get_singles(n_orbs, up_down=up_down)
    doubles = get_doubles(n_orbs, up_down=up_down)
    return singles + doubles


def uccgsd_generator(n_qubits, single_coeffs=None, double_coeffs=None, up_down=False):
    """Construct a list of FermionOperators enumerating the Unitary
    Coupled-Cluster Generalized Singles and Doubles excitations. Groups of
    spin-compensating and conjugate terms are combined together, with their
    relative weights to enforce spin-symmetries and reduce redundancy in VQE
    optimization parameters. All singles are followed by double excitations.

    Args:
        n_qubits (int): number of qubits (spin-orbitals) in basis (this is
            number of basis-orbitals multiplied by 2).
        up_down (bool): flag for using qiskit (True) or openfermion (False)
            spin-ordering conventions.

    Returns:
        all_operators (list): list of FermionOperator objects.
    """
    if type(n_qubits) != int:
        raise TypeError("Number of qubits (n_qubits) must be integer type.")
    if n_qubits < 2:
        raise ValueError("Number of qubits (n_qubits) must be at least 2.")
    elif n_qubits % 2 != 0:
        raise ValueError("Invalid number of qubits (n_qubits) -- must be even.")
    if type(up_down) != bool:
        raise TypeError("Spin-ordering arg (up_down) must be boolean.")

    coeffs = get_coeffs(n_qubits, single_coeffs, double_coeffs)  # check coeffecients passed, or generate random ones
    operators = get_all_excitations(n_qubits // 2, up_down=False)  # get all operator input arguments

    all_operators = list()
    for index, oi in enumerate(operators):
        # create new FermionOperator
        current = FermionOperator()
        for term in oi:
            # add term to operator
            current += FermionOperator(*term[:-1], term[-1] * coeffs[index])
        all_operators.append(current)

    return all_operators


def get_singles_number(n_orbitals):
    """Get number of independent terms in the set of singles excitations for a
    designated number of orbital states. Note argument is the number of spatial
    orbital basis states, number of spin orbitals is twice as large.

    Args:
        n_orbitals(int): number of basis orbitals.

    Return:
        number of singles excitations (int).
    """
    if not isinstance(n_orbitals, int):
        raise TypeError("Number of orbitals (n_orbitals) must be integer valued")
    return n_orbitals * (n_orbitals - 1) // 2


def get_doubles_number(n_orbitals):
    """Get number of independent terms in the set of doubles excitations for a
    designated number of orbital states. Note argument is the number of spatial
    orbital basis states, number of spin orbitals is twice as large.

    Args:
        n_orbitals(int): number of basis orbitals.

    Return:
        number of doubles excitations (int).
    """
    if not isinstance(n_orbitals, int):
        raise TypeError("Number of orbitals (n_orbitals) must be an integer")
    return n_orbitals * (n_orbitals ** 3 + 2 * n_orbitals ** 2 - n_orbitals - 2) // 8


def get_excitation_number(n_orbitals):
    """Get number of independent singles and doubles excitations for a given
    number of basis orbital (spatial) states. Note input argument is half the
    number of spin-orbitals.

    Args:
        n_orbitals(int): number of basis orbitals.

    Return:
        number of excitations (int).
    """

    return get_singles_number(n_orbitals) + get_doubles_number(n_orbitals)


def get_coeffs(n_qubits, single_coeffs=None, double_coeffs=None):
    """Prepare coefficients for UCCGSD excitation terms. User has option to
    either pass specified singles and doubles excitation coefficients, or to
    utilize randomly chosen values. Note, if passing own values, check that
    number is consistent with the output for *get_singles_number* and
    *get_doubles_number* before proceeding.

    Args:
        n_qubits (int): number of qubits for problem (2x number of orbitals).
        single_coeffs (list or numpy array, OR None): initial singles-excitation
            coefficients, or skip.
        double_coeffs (list or numpy array, OR None)): initial doubles-excitation
            coefficients, or skip.
    Returns:
        coeffs (numpy array of float): coefficients for initializing the
            FermionOperator for the UCCGSD.
    """
    if single_coeffs is None:
        single_coeffs = np.random.random(get_singles_number(n_qubits // 2))
    elif len(single_coeffs) != get_singles_number(n_qubits // 2):
        raise ValueError(
            f"Invalid number of single excitation coefficients, expecting {get_singles_number(n_qubits // 2)}.")

    if double_coeffs is None:
        double_coeffs = np.random.random(get_doubles_number(n_qubits // 2))
    elif len(double_coeffs) != get_doubles_number(n_qubits // 2):
        raise ValueError(
            f"Invalid number of double excitation coefficients, expecting {get_doubles_number(n_qubits // 2)}.")

    coeffs = np.concatenate((single_coeffs, double_coeffs))

    return coeffs



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module defining a class to store objects related to qubit tapering, i.e.
resource reduction through symmetries.
"""

import numpy as np

from tangelo.toolboxes.operators import MultiformOperator
from tangelo.toolboxes.operators.z2_tapering import get_clifford_operators, get_unitary, get_eigenvalues, get_z2_taper_function


class QubitTapering:
    """Class keeping track of symmetries and operation to taper qubits in qubit
    operators. The taper function is kept as an attribute to taper other Pauli
    word generators (Hamiltonian dressing, ansatz construction, etc). Z2
    tapering is implemented, but other tapering methods could be added in the
    core of this class.

    Attributes:
        initial_op (MultiformOperator): Qubit operator to be analyzed for
            symmetries.
        initial_n_qubits (int): Number of qubits before tapering.
        z2_taper (func): Function handle for tapering a MultiformOperator.
        z2_tapered_op (MultiformOperator): Tapered operator with z2 symmetries.
        z2_properties (dict): Relevant quantities used to define the z2 taper
            function. Needed to back track a z2 tapered operator to the full
            operator.
    """

    def __init__(self, qubit_operator, n_qubits, n_electrons, spin=0, mapping="JW", up_then_down=False):
        """Class keeping track of symmetries and operation to taper qubits in
        qubit operators.

        Args:
            qubit_operator (QubitOperator): Self-explanatory.
            n_qubits (int): Number of qubits the initial operator acts on.
            n_electrons (int): Number of electrons.
            mapping (str): Qubit mapping.
            up_then_down (bool): Whether or not spin ordering is all up then
                all down.
        """

        if mapping.upper() not in {"JW", "BK", "JKMN"}:
            raise NotImplementedError(f"Qubit mapping {mapping} not supported. Tapering supports JW, BK and JKMN qubit encoding.")

        self.initial_op = MultiformOperator.from_qubitop(qubit_operator, n_qubits)
        self.initial_n_qubits = n_qubits
        self.n_electrons = n_electrons
        self.spin = spin
        self.mapping = mapping
        self.up_then_down = up_then_down

        self.z2_taper = None
        self.z2_tapered_op = None
        self.z2_properties = dict()

        self._compute_z2_symmetries()

    def z2_tapering(self, qubit_operator, n_qubits=None):
        """Function to taper a qubit operator from symmetries found in
        self.initial_op.

        Args:
            qubit_operator (QubitOperator): Self-explanatory.
            n_qubits (int): Self-explanatory.

        Returns:
            QubitOperator: The tapered qubit operator.
        """

        hybrid_op = MultiformOperator.from_qubitop(qubit_operator, n_qubits)
        z2_tapered_op = self.z2_taper(hybrid_op)

        return z2_tapered_op.qubitoperator

    def _compute_z2_symmetries(self):
        """Computes the underlying z2 symmetries in elf.initial_op. The
        procedure is described in
        Tapering off qubits to simulate fermionic Hamiltonians
        Sergey Bravyi, Jay M. Gambetta, Antonio Mezzacapo, Kristan Temme.
        arXiv:1701.08213
        """

        kernel = self.initial_op.get_kernel()
        cliffords, q_indices = get_clifford_operators(kernel)
        n_symmetries = len(q_indices)
        unitary = get_unitary(cliffords)

        kernel_operator = MultiformOperator.from_binaryop(kernel, factors=np.ones(kernel.shape[0]))
        eigenvalues = get_eigenvalues(kernel_operator.binary, self.initial_n_qubits, self.n_electrons, self.spin, self.mapping, self.up_then_down)

        self.z2_taper = get_z2_taper_function(unitary, kernel_operator, q_indices, self.initial_n_qubits, n_symmetries, eigenvalues)
        self.z2_tapered_op = self.z2_taper(self.initial_op)
        self.z2_properties = {"n_symmetries": n_symmetries, "eigenvalues": eigenvalues, "unitary": unitary}



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines the qubit mean-field (QMF) ansatz class. The ansatz is a
variational product state built from a set of parameterized single-qubit states.
For applications in quantum chemistry, the ansatz can be used to describe the
mean-field component of an electronic wave function on a quantum computer.
For more information about this ansatz, see references below.

Refs:
    1. I. G. Ryabinkin, T.-C. Yen, S. N. Genin, and A. F. Izmaylov.
        J. Chem. Theory Comput. 2018, 14 (12), 6317-6326.
    2. I. G. Ryabinkin and S. N. Genin.
        https://arxiv.org/abs/1812.09812 2018.
    3. S. N. Genin, I. G. Ryabinkin, and A. F. Izmaylov.
          https://arxiv.org/abs/1901.04715 2019.
    4. I. G. Ryabinkin, S. N. Genin, and A. F. Izmaylov.
        J. Chem. Theory Comput. 2019, 15, 1, 249–255.
"""

import warnings

import numpy as np

from tangelo.toolboxes.qubit_mappings.mapping_transform import get_qubit_number,\
                                                               fermion_to_qubit_mapping
from tangelo.toolboxes.ansatz_generator.ansatz import Ansatz
from tangelo.toolboxes.ansatz_generator._qubit_mf import get_qmf_circuit, init_qmf_from_hf,\
                                                         penalize_mf_ham, init_qmf_from_vector


class QMF(Ansatz):
    """This class implements the QMF ansatz. Closed-shell and restricted open-shell QMF are
    supported. While the form of the QMF ansatz is the same for either variation, the underlying
    fermionic mean-field state is treated differently depending on the spin. Closed-shell
    or restricted open-shell QMF implies that spin = 0 or spin != 0 and the fermionic mean-field
    state is obtained using a RHF or ROHF Hamiltonian, respectively.

    Optimizing QMF variational parameters can be risky without taking proper precautions,
    especially when a random initial guess is used. It is recommended that penalty terms are
    added to the mean-field Hamiltonian to enforce appropriate electron number and spin angular
    momentum symmetries on the QMF wave function during optimization (see Ref. 4). If using
    penalty terms is to be avoided, an initial guess based on a Hartree-Fock reference state will
    likely converge quickly to the desired state, but this is not guaranteed.

    Args:
        molecule (SecondQuantizedMolecule): The molecular system.
        mapping (str): One of the supported qubit mapping identifiers. Default, "jw".
        up_then_down (bool): Change basis ordering putting all spin up orbitals first,
            followed by all spin down. Default, False.
        init_qmf (dict): Controls for QMF variational parameter initialization and mean-field
            Hamiltonian penalization. Supported keys are "init_params", "N", "S^2", or "Sz" (str).
            Values of "init_params" must be in self.supported_initial_var_params (str). Values of
            "N", "S^2", or "Sz" are (tuple or None). If tuple, the elements are a penalty
            term coefficient, mu (float), and a target value of the penalty operator (int).
            Example - "key": (mu, target). If "N", "S^2", or "Sz" is None, a penalty term is added
            with default mu and target values: mu = 1.5 and target is derived from molecule as
            <N> = n_electrons, <S^2> = spin_z * (spin_z + 1), and <Sz> = spin_z, where
            spin_z = spin // 2. Key, value pairs are case sensitive and mu > 0.
            Default, {"init_params": "hf_state"}.
    """

    def __init__(self, molecule, mapping="jw", up_then_down=False, init_qmf=None):

        if not molecule:
            raise ValueError("An instance of SecondQuantizedMolecule is required for initializing "
                             "the self.__class__.__name__ ansatz class.")
        self.molecule = molecule

        self.n_spinorbitals = self.molecule.n_active_sos
        if self.n_spinorbitals % 2 != 0:
            raise ValueError("The total number of spin-orbitals should be even.")

        self.n_orbitals = self.n_spinorbitals // 2
        self.spin = molecule.spin
        self.fermi_ham = self.molecule.fermionic_hamiltonian
        self.n_electrons = self.molecule.n_active_electrons

        self.mapping = mapping
        self.n_qubits = get_qubit_number(self.mapping, self.n_spinorbitals)
        self.up_then_down = up_then_down
        self.init_qmf = {"init_params": "hf_state"} if init_qmf is None else init_qmf

        # Supported var param initialization
        self.supported_initial_var_params = {"vacuum", "half_pi", "minus_half_pi", "full_pi",
                                             "random", "hf_state", "vector"}

        # Supported reference state initialization
        self.supported_reference_state = {"HF"}

        # Add any penalty terms to the fermionic Hamiltonian
        if isinstance(self.init_qmf, dict):
            if "init_params" not in self.init_qmf.keys():
                raise KeyError(f"Missing key 'init_params' in {self.init_qmf}. "
                               f"Supported values are {self.supported_initial_var_params}.")
            if self.init_qmf["init_params"] in self.supported_initial_var_params:
                # Set the default QMF parameter procedure
                self.var_params_default = self.init_qmf.pop("init_params")
                if self.var_params_default == "vector":
                    self.vector = self.init_qmf.pop("vector")
                # Check for at least one penalty term
                if self.init_qmf:
                    # Set default penalty term values
                    spin_z = self.spin // 2
                    init_qmf_defaults = {"N": (1.5, self.n_electrons),
                                         "S^2": (1.5, spin_z * (spin_z + 1)), "Sz": (1.5, spin_z)}
                    # Check if the user requested default values for any penalty terms
                    for term, params in init_qmf_defaults.items():
                        if params is None:
                            self.init_qmf[term] = init_qmf_defaults[term]
                    # Add the penalty terms to the mean-field Hamiltonian
                    self.fermi_ham += penalize_mf_ham(self.init_qmf, self.n_orbitals)
                else:
                    if self.var_params_default != "hf_state":
                        warnings.warn("It is recommended that the QMF parameters are initialized "
                                      "using a Hartree-Fock reference state if penalty terms are "
                                      "not added to the mean-field Hamiltonian.", RuntimeWarning)
            else:
                raise ValueError(f"Unrecognized value for 'init_params' key in {self.init_qmf} "
                                 f"Supported values are {self.supported_initial_var_params}.")
        else:
            raise TypeError(f"{self.init_qmf} must be dictionary type.")

        # Get the qubit Hamiltonian
        self.qubit_ham = fermion_to_qubit_mapping(self.fermi_ham, self.mapping, self.n_spinorbitals,
                                                  self.n_electrons, self.up_then_down, self.spin)

        # Default starting parameters for initialization
        self.n_var_params = 2 * self.n_qubits
        self.default_reference_state = "HF"
        self.var_params = None
        self.circuit = None

    def set_var_params(self, var_params=None):
        """Set values for variational parameters, such as zeros, random numbers,
        or a Hartree-Fock state occupation vector, providing some keywords
        for users, and also supporting direct user input (list or numpy array).
        Return the parameters so that workflows such as VQE can retrieve these values. """

        if var_params is None:
            var_params = self.var_params_default

        if isinstance(var_params, str):
            var_params = var_params.lower()
            if var_params not in self.supported_initial_var_params:
                raise ValueError(f"Supported keywords for initializing variational parameters: "
                                 f"{self.supported_initial_var_params}")
            # Initialize |QMF> as |00...0>
            if var_params == "vacuum":
                initial_var_params = np.zeros((self.n_var_params,), dtype=float)
            # Initialize |QMF> as (1/sqrt(2))^n_qubits * tensor_prod(|0> + 1j|1>)
            elif var_params == "half_pi":
                initial_var_params = 0.5 * np.pi * np.ones((self.n_var_params,))
            # Initialize |QMF> as (1/sqrt(2))^n_qubits * tensor_prod(|0> - 1j|1>)
            elif var_params == "minus_half_pi":
                initial_var_params = -0.5 * np.pi * np.ones((self.n_var_params,))
            # Initialize |QMF> as (i)^n_qubits * |11...1>
            elif var_params == "full_pi":
                initial_var_params = np.pi * np.ones((self.n_var_params,))
            # Random initialization of thetas over [0, pi] and phis over [0, 2 * pi]
            elif var_params == "random":
                initial_thetas = np.pi * np.random.random((self.n_qubits,))
                initial_phis = 2. * np.pi * np.random.random((self.n_qubits,))
                initial_var_params = np.concatenate((initial_thetas, initial_phis))
            # Initialize thetas as 0 or pi such that |QMF> = |HF> and set all phis to 0
            elif var_params == "hf_state":
                initial_var_params = init_qmf_from_hf(self.n_spinorbitals, self.n_electrons,
                                                      self.mapping, self.up_then_down, self.spin)
            elif var_params == "vector":
                initial_var_params = init_qmf_from_vector(self.vector, self.mapping, self.up_then_down)
        else:
            initial_var_params = np.array(var_params)
            if initial_var_params.size != self.n_var_params:
                raise ValueError(f"Expected {self.n_var_params} variational parameters but "
                                 f"received {initial_var_params.size}.")
        self.var_params = initial_var_params
        return initial_var_params

    def prepare_reference_state(self):
        """Returns circuit preparing the reference state of the ansatz (e.g prepare reference
        wavefunction with HF, multi-reference state, etc). These preparations must be consistent
        with the transform used to obtain the qubit operator. """

        if self.default_reference_state not in self.supported_reference_state:
            raise ValueError(f"Only supported reference state methods are: "
                             f"{self.supported_reference_state}")
        if self.default_reference_state == "HF":
            reference_state_circuit = get_qmf_circuit(self.var_params, True)
        return reference_state_circuit

    def build_circuit(self, var_params=None):
        """Build and return the quantum circuit implementing the state preparation ansatz
         (with currently specified initial_state and var_params). """

        if var_params is not None:
            self.set_var_params(var_params)
        elif self.var_params is None:
            self.set_var_params()

        # Build the circuit for the QMF ansatz
        self.circuit = self.prepare_reference_state()

    def update_var_params(self, var_params):
        """Shortcut: set value of variational parameters in the already-built ansatz circuit member.
        Preferable to rebuilt your circuit from scratch, which can be an involved process. """

        # Rebuild the circuit because it is trivial
        self.build_circuit(var_params)



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import numpy as np

from tangelo.algorithms import BuiltInAnsatze, SA_VQESolver
from tangelo.molecule_library import mol_H2_sto3g
from tangelo.toolboxes.ansatz_generator import UCCSD
from tangelo.toolboxes.qubit_mappings.mapping_transform import fermion_to_qubit_mapping
from tangelo.toolboxes.molecular_computation.rdms import matricize_2rdm


class SA_VQESolverTest(unittest.TestCase):

    def test_instantiation_sa_vqe(self):
        """Try instantiating SA_VQESolver with basic input."""

        options = {"molecule": mol_H2_sto3g, "qubit_mapping": "jw", "ref_states": [[1, 1, 0, 0]]}
        SA_VQESolver(options)

    def test_instantiation_sa_vqe_incorrect_keyword(self):
        """Instantiating with an incorrect keyword should return an error """

        options = {"molecule": mol_H2_sto3g, "qubit_mapping": "jw", "dummy": True}
        self.assertRaises(KeyError, SA_VQESolver, options)

    def test_instantiation_sa_vqe_missing_molecule(self):
        """Instantiating with no molecule should return an error."""

        options = {"qubit_mapping": "jw"}
        self.assertRaises(ValueError, SA_VQESolver, options)

    def test_instantiation_sa_vqe_missing_ref_states(self):
        """Instantiating with no ref_states should return an error."""

        options = {"molecule": mol_H2_sto3g, "qubit_mapping": "jw"}
        self.assertRaises(ValueError, SA_VQESolver, options)

    def test_get_resources_h2_mappings(self):
        """Resource estimation, with UCCSD ansatz, given initial parameters.
        Each of JW, BK, and scBK mappings are checked.
        """
        mappings = ["jw", "bk", "scbk", "jkmn"]
        expected_values = [(15, 4), (15, 4), (5, 2), (15, 4)]

        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": "jw",
                       "initial_var_params": [0.1, 0.1], "ref_states": [[1, 1, 0, 0]]}
        for index, mi in enumerate(mappings):
            vqe_options["qubit_mapping"] = mi
            sa_vqe_solver = SA_VQESolver(vqe_options)
            sa_vqe_solver.build()
            resources = sa_vqe_solver.get_resources()

            self.assertEqual(resources["qubit_hamiltonian_terms"], expected_values[index][0])
            self.assertEqual(resources["circuit_width"], expected_values[index][1])

    def test_energy_estimation_sa_vqe(self):
        """A single SA-VQE energy evaluation for H2, using optimal parameters and
        exact simulator.
        """

        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": "jw",
                       "ref_states": [[1, 1, 0, 0]]}
        sa_vqe_solver = SA_VQESolver(vqe_options)
        sa_vqe_solver.build()

        energy = sa_vqe_solver.energy_estimation([5.86665842e-06, 5.65317429e-02])
        self.assertAlmostEqual(energy, -1.137270422018, places=6)

    def test_operator_expectation_sa_vqe(self):
        """ A test of the operator_expectation function, using optimal parameters and exact simulator """

        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": 'jw', "ref_states": [[1, 1, 0, 0]]}
        sa_vqe_solver = SA_VQESolver(vqe_options)
        sa_vqe_solver.build()

        # Test using var_params input and Qubit Hamiltonian
        energy = sa_vqe_solver.operator_expectation(sa_vqe_solver.qubit_hamiltonian, var_params=[5.86665842e-06, 5.65317429e-02],
                                                    ref_state=sa_vqe_solver.reference_circuits[0])
        self.assertAlmostEqual(energy, -1.137270422018, places=6)

        # Test using updated var_params and Fermion Hamiltonian
        sa_vqe_solver.ansatz.update_var_params([5.86665842e-06, 5.65317429e-02])
        energy = sa_vqe_solver.operator_expectation(mol_H2_sto3g.fermionic_hamiltonian, ref_state=sa_vqe_solver.reference_circuits[0])
        self.assertAlmostEqual(energy, -1.137270422018, places=6)

        # Test the three in place operators
        n_electrons = sa_vqe_solver.operator_expectation('N', ref_state=sa_vqe_solver.reference_circuits[0])
        self.assertAlmostEqual(n_electrons, 2, places=6)
        spin_z = sa_vqe_solver.operator_expectation('Sz', ref_state=sa_vqe_solver.reference_circuits[0])
        self.assertAlmostEqual(spin_z, 0, places=6)
        spin2 = sa_vqe_solver.operator_expectation('S^2', ref_state=sa_vqe_solver.reference_circuits[0])
        self.assertAlmostEqual(spin2, 0, places=6)

    def test_simulate_h2(self):
        """Run SA-VQE on H2 molecule, with UpCCGSD ansatz, JW qubit mapping, ref_states and exact simulator.
        Followed by deflation of two states calculated to determine next excited state.
        """

        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UpCCGSD, "qubit_mapping": "jw",
                       "verbose": True, "ref_states": [[1, 1, 0, 0], [1, 0, 1, 0]]}
        sa_vqe_solver = SA_VQESolver(vqe_options)
        sa_vqe_solver.build()

        # code to generate exact results.
        # from pyscf import mcscf
        # mc = mcscf.CASCI(mol_H2_sto3g.mean_field, 2, 2)
        # mc.fcisolver.nroots = 3
        # exact_energies = mc.casci()[0]
        exact_energies = [-1.1372702, -0.5324790, -0.1699013]

        # Use state averaging to get ground and first excited state.
        _ = sa_vqe_solver.simulate()
        np.testing.assert_array_almost_equal(exact_energies[:2], sa_vqe_solver.state_energies, decimal=3)

        # Use deflation to get second excited state from circuits for ground state and first excited state
        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UpCCGSD, "qubit_mapping": "jw",
                       "verbose": True, "deflation_circuits": [sa_vqe_solver.reference_circuits[0] + sa_vqe_solver.optimal_circuit,
                                                               sa_vqe_solver.reference_circuits[1] + sa_vqe_solver.optimal_circuit],
                       "deflation_coeff": 1.5, "ref_states": [[0, 0, 1, 1]]}
        vqe_solver_2 = SA_VQESolver(vqe_options)
        vqe_solver_2.build()

        energy = vqe_solver_2.simulate()
        self.assertAlmostEqual(exact_energies[2], energy, delta=1.e-3)

    def test_get_rdm_h2(self):
        """Compute RDMs with UCCSD ansatz, JW qubit mapping, optimized
        parameters, exact simulator (H2).
        """

        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": "jw", "ref_states": [[1, 1, 0, 0]]}
        sa_vqe_solver = SA_VQESolver(vqe_options)
        sa_vqe_solver.build()

        # Compute RDM matrices
        one_rdm, two_rdm = sa_vqe_solver.get_rdm([5.86665842e-06, 5.65317429e-02], ref_state=sa_vqe_solver.reference_circuits[0])

        # Test traces of matrices
        n_elec, n_orb = mol_H2_sto3g.n_active_electrons, mol_H2_sto3g.n_active_mos
        self.assertAlmostEqual(np.trace(one_rdm), n_elec, msg="Trace of one_rdm does not match number of electrons",
                               delta=1e-6)
        rho = matricize_2rdm(two_rdm, n_orb)
        self.assertAlmostEqual(np.trace(rho), n_elec * (n_elec - 1),
                               msg="Trace of two_rdm does not match n_elec * (n_elec-1)", delta=1e-6)

    def test_custom_vqe(self):
        """SA-VQE with custom optimizer and non-optimal variational parameters."""

        # Define and assign custom optimizer: cobyla
        def cobyla_oneshot_optimizer(func, var_params):
            from scipy.optimize import minimize
            result = minimize(func, var_params, method="COBYLA", options={"disp": True, "maxiter": 100})
            return result.fun, result.x

        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": "jw",
                       "initial_var_params": "ones", "verbose": False, "ref_states": [[1, 1, 0, 0]],
                       "optimizer": cobyla_oneshot_optimizer}
        sa_vqe_solver = SA_VQESolver(vqe_options)
        sa_vqe_solver.build()

        energy = sa_vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.137270422018, places=6)

    def test_qubit_qhamiltonian_input(self):
        """Test the case where a qubit Hamiltonian is used to construct SA-VQE."""

        qubit_hamiltonian = fermion_to_qubit_mapping(mol_H2_sto3g.fermionic_hamiltonian, mapping="jw")

        options = {"qubit_hamiltonian": qubit_hamiltonian,
                   "ansatz": UCCSD(mol_H2_sto3g, mapping="jw"), "ref_states": [[1, 1, 0, 0]]}
        SA_VQESolver(options)

    def test_qubit_qhamiltonian_input_conflicts(self):
        """Test the case where a molecule and a qubit Hamiltonian are passed as
        inputs.
        """

        qubit_hamiltonian = fermion_to_qubit_mapping(mol_H2_sto3g.fermionic_hamiltonian, mapping="jw")

        options = {"molecule": mol_H2_sto3g,
                   "qubit_hamiltonian": qubit_hamiltonian,
                   "ansatz": UCCSD(mol_H2_sto3g, mapping="jw"), "ref_states": [[1, 1, 0, 0]]}

        with self.assertRaises(ValueError):
            SA_VQESolver(options)

    def test_qubit_qhamiltonian_input_no_custom_ansatz(self):
        """Test the case where no custom ansatz is passed when using a qubit
        Hamiltonian as input.
        """

        qubit_hamiltonian = fermion_to_qubit_mapping(mol_H2_sto3g.fermionic_hamiltonian, mapping="jw")

        options = {"qubit_hamiltonian": qubit_hamiltonian, "ref_states": [[1, 1, 0, 0]]}

        with self.assertRaises(TypeError):
            SA_VQESolver(options).build()

        options["ansatz"] = BuiltInAnsatze.UCCSD

        with self.assertRaises(TypeError):
            SA_VQESolver(options).build()


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import os
import numpy as np

from tangelo.toolboxes.post_processing import mcweeny_purify_2rdm

path_data = os.path.dirname(os.path.abspath(__file__)) + "/data"


class McWeenyPurificationTest(unittest.TestCase):

    def test_mcweeny_exact(self):
        """Check McWeeny"s RDM purification technique using exact simulation
        results. Input, output and code as used in
        https://arxiv.org/pdf/2102.07045.pdf, R=1.1 A (pot 7) of the H10 curve.
        """
        with open(f"{path_data}/rdm2_spin_pot7_exact.npy", "rb") as f:
            rdm2_spin = np.load(f)
        with open(f"{path_data}/rdm1_nw_exact.npy", "rb") as f:
            rdm1_mw_ref = np.load(f)
        with open(f"{path_data}/rdm2_nw_exact.npy", "rb") as f:
            rdm2_mw_ref = np.load(f)

        rdm1_mw, rdm2_mw = mcweeny_purify_2rdm(rdm2_spin)
        np.testing.assert_array_almost_equal(rdm1_mw, rdm1_mw_ref)
        np.testing.assert_array_almost_equal(rdm2_mw, rdm2_mw_ref)

    def test_mcweeny_experimental(self):
        """ Check McWeeny"s RDM purification technique using experimental
        simulation results. Input, output and code as used in
        https://arxiv.org/pdf/2102.07045.pdf, R=1.1 A (pot 7) of the H10 curve.
        """
        with open(f"{path_data}/rdm2_spin_pot7_experimental.npy", "rb") as f:
            rdm2_spin = np.load(f)
        with open(f"{path_data}/rdm1_nw.npy", "rb") as f:
            rdm1_mw_ref = np.load(f)
        with open(f"{path_data}/rdm2_nw.npy", "rb") as f:
            rdm2_mw_ref = np.load(f)

        rdm1_mw, rdm2_mw = mcweeny_purify_2rdm(rdm2_spin)
        np.testing.assert_array_almost_equal(rdm1_mw, rdm1_mw_ref)
        np.testing.assert_array_almost_equal(rdm2_mw, rdm2_mw_ref)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions helping with quantum circuit format conversion between abstract
format and a subset of the openqasm format (the format generated by IBM openqasm
functionalities in qiskit).

In order to produce an equivalent circuit for the target backend, it is
necessary to account for:
- how the gate names differ between the source backend to the target backend.
- how the order and conventions for some of the inputs to the gate operations
    may also differ.
"""

import re
from math import pi
from tangelo.linq import Gate, Circuit


def get_openqasm_gates():
    """Map gate name of the abstract format to the equivalent gate name used in
    openqasm OpenQASM is a general format that allows users to express a quantum
    program, define conditional operations manipulating quantum and qubit
    registers, as well as defining new quantum unitaries. We however make the
    choice here to support well-known gate operations.
    """

    GATE_OPENQASM = dict()
    for name in {"H", "X", "Y", "Z", "S", "T", "RX", "RY", "RZ", "MEASURE",
                 "CZ", "CY", "CRZ", "SWAP", "CSWAP"}:
        GATE_OPENQASM[name] = name.lower()
    GATE_OPENQASM["CNOT"] = "cx"
    GATE_OPENQASM["PHASE"] = "p"
    GATE_OPENQASM["CPHASE"] = "cp"

    return GATE_OPENQASM


def translate_openqasm(source_circuit):
    """Take in an abstract circuit, return a OpenQASM 2.0 string using IBM
    Qiskit (they are the reference for OpenQASM).

    Args:
        source_circuit: quantum circuit in the abstract format.

    Returns:
        str: the corresponding OpenQASM program, as per IBM Qiskit.
    """
    from .translate_qiskit import translate_qiskit

    return translate_qiskit(source_circuit).qasm()


def _translate_openqasm2abs(openqasm_str):
    """Take an OpenQASM 2.0 string as input (as defined by IBM Qiskit), return
    the equivalent abstract circuit. Only a subset of OpenQASM supported, mostly
    to be able to go back and forth QASM and abstract representations to
    leverage tools and innovation implemented to work in the QASM format. Not
    designed to support elaborate QASM programs defining their own operations.
    Compatible with qiskit.QuantumCircuit.from_qasm method.

    Assumes single-qubit measurement instructions only. Final qubit register
    measurement is implicit.

    Args:
        openqasm_string(str): an OpenQASM program, as a string, as defined by
            IBM Qiskit.

    Returns:
        Circuit: corresponding quantum circuit in the abstract format.
    """

    # Get dictionary of gate mapping, as the reverse dictionary of abs -> openqasm translation
    GATE_OPENQASM = get_openqasm_gates()
    gate_mapping = {v: k for k, v in GATE_OPENQASM.items()}

    def parse_param(s):
        """ Parse parameter as either a float or a string if it's not a float """
        try:
            return float(s)
        except ValueError:
            return s

    # Get number of qubits, extract gate operations
    n_qubits = int(re.findall('qreg q\[(\d+)\];', openqasm_str)[0])
    openqasm_gates = openqasm_str.split(f"qreg q[{n_qubits}];\ncreg c[{n_qubits}];")[-1]
    openqasm_gates = [instruction for instruction in openqasm_gates.split("\n") if instruction]

    # Translate gates
    abs_circ = Circuit()
    for openqasm_gate in openqasm_gates:

        # Extract gate name, qubit indices and parameter value (single parameter for now)
        gate_name = re.split('\s|\(', openqasm_gate)[0]
        qubit_indices = [int(index) for index in re.findall('q\[(\d+)\]', openqasm_gate)]
        parameters = [parse_param(index) for index in re.findall('\((.*)\)', openqasm_gate)]
        # TODO: controlled operation, will need to store value in classical register
        #  bit_indices = [int(index) for index in re.findall('c\[(\d+)\]', openqasm_gate)]

        if gate_name in {"h", "x", "y", "z", "s", "t", "measure"}:
            gate = Gate(gate_mapping[gate_name], qubit_indices[0])
        elif gate_name in {"rx", "ry", "rz", "p"}:
            gate = Gate(gate_mapping[gate_name], qubit_indices[0], parameter=eval(str(parameters[0])))
        # TODO: Rethink the use of enums for gates to set the equality CX=CNOT and enable other refactoring
        elif gate_name in {"cx", "cz", "cy"}:
            gate = Gate(gate_mapping[gate_name], qubit_indices[1], control=qubit_indices[0])
        elif gate_name in {"crz", "cp"}:
            gate = Gate(gate_mapping[gate_name], qubit_indices[1], control=qubit_indices[0], parameter=eval(str(parameters[0])))
        elif gate_name in {"swap"}:
            gate = Gate(gate_mapping[gate_name], [qubit_indices[0], qubit_indices[1]])
        elif gate_name in {"cswap"}:
            gate = Gate(gate_mapping[gate_name], [qubit_indices[1], qubit_indices[2]], control=qubit_indices[0])
        else:
            raise ValueError(f"Gate '{gate_name}' not supported with openqasm translation")
        abs_circ.add_gate(gate)

    return abs_circ



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for the closed-shell and restricted open-shell iQCC-VQE Solver. """

import unittest

from tangelo.algorithms.variational import iQCC_solver
from tangelo.molecule_library import mol_H2_sto3g, mol_H4_sto3g, mol_H4_cation_sto3g,\
                                     mol_H4_doublecation_minao


class iQCC_solver_test(unittest.TestCase):
    """Unit tests for the iQCC_solver class. Examples for both closed-shell
    and restricted open-shell iQCC are provided via H4, H4+, and H4+2.
    """

    @staticmethod
    def test_build_success():
        """Test instantation of iQCC solver with user-defined input."""

        iqcc_options = {"molecule": mol_H2_sto3g,
                        "qubit_mapping": "scbk",
                        "up_then_down": True,
                        "deqcc_thresh": 1e-5,
                        "max_iqcc_iter": 25,
                        "max_iqcc_retries": 10,
                        "compress_qubit_ham": True,
                        "compress_eps": 1e-4}

        iqcc = iQCC_solver(iqcc_options)
        iqcc.build()

    def test_build_fail(self):
        """Test that instantation of iQCC solver fails without input of a molecule."""

        iqcc_options = {"max_iqcc_iter": 15}
        self.assertRaises(ValueError, iQCC_solver, iqcc_options)

    def test_iqcc_h4(self):
        """Test the energy after 1 iteration for H4 using the maximum
        number of generators and compressing the qubit Hamiltonian"""

        ansatz_options = {"max_qcc_gens": None}

        iqcc_options = {"molecule": mol_H4_sto3g,
                        "qubit_mapping": "scbk",
                        "up_then_down": True,
                        "ansatz_options": ansatz_options,
                        "deqcc_thresh": 1e-5,
                        "max_iqcc_iter": 1,
                        "compress_qubit_ham": True,
                        "compress_eps": 1e-4}

        iqcc_solver = iQCC_solver(iqcc_options)
        iqcc_solver.build()
        iqcc_energy = iqcc_solver.simulate()

        self.assertAlmostEqual(iqcc_energy, -1.96259, places=4)

    def test_iqcc_h4_cation(self):
        """Test the energy after 3 iterations for H4+"""

        ansatz_options = {"max_qcc_gens": None}

        iqcc_options = {"molecule": mol_H4_cation_sto3g,
                        "qubit_mapping": "scbk",
                        "up_then_down": True,
                        "ansatz_options": ansatz_options,
                        "deqcc_thresh": 1e-5,
                        "max_iqcc_iter": 3}

        iqcc = iQCC_solver(iqcc_options)
        iqcc.build()
        iqcc_energy = iqcc.simulate()

        self.assertAlmostEqual(iqcc_energy, -1.638524, places=4)

    def test_iqcc_h4_double_cation(self):
        """Test the energy after 1 iteration for H4+2"""

        ansatz_options = {"max_qcc_gens": None}

        iqcc_options = {"molecule": mol_H4_doublecation_minao,
                        "qubit_mapping": "scbk",
                        "up_then_down": True,
                        "ansatz_options": ansatz_options,
                        "deqcc_thresh": 1e-5,
                        "max_iqcc_iter": 1}

        iqcc = iQCC_solver(iqcc_options)
        iqcc.build()
        iqcc_energy = iqcc.simulate()

        self.assertAlmostEqual(iqcc_energy, -0.854647, places=4)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Simple tests for input to mapping_transform.py. Presently, mostly using
wrappers on openfermion code, so pretty minimal testing here.
"""

import unittest
import numpy as np

from openfermion.linalg import eigenspectrum
from openfermion.linalg.sparse_tools import qubit_operator_sparse

from tangelo.toolboxes.operators import QubitOperator, FermionOperator
from tangelo.toolboxes.qubit_mappings.mapping_transform import fermion_to_qubit_mapping, make_up_then_down


class MappingTest(unittest.TestCase):

    def test_bk(self):
        """Check output from Bravyi-Kitaev transformation."""
        bk_operator = QubitOperator(((0, "Z"), (1, "Y"), (2, "X")), -0.25j)
        bk_operator += QubitOperator(((0, "Z"), (1, "Y"), (2, "Y")), -0.25)
        bk_operator += QubitOperator(((1, "X"), (2, "X")), -0.25)
        bk_operator += QubitOperator(((1, "X"), (2, "Y")), 0.25j)
        bk_operator += QubitOperator(((0, "X"), (1, "Y"), (2, "Z")), -0.125j)
        bk_operator += QubitOperator(((0, "X"), (1, "X"), (3, "Z")), -0.125)
        bk_operator += QubitOperator(((0, "Y"), (1, "Y"), (2, "Z")), -0.125)
        bk_operator += QubitOperator(((0, "Y"), (1, "X"), (3, "Z")), 0.125j)

        fermion = FermionOperator(((1, 0), (2, 1)), 1.0) + FermionOperator(((0, 1), (3, 0)), 0.5)
        n_spinorbitals = 4
        qubit = fermion_to_qubit_mapping(fermion, mapping="BK", n_spinorbitals=n_spinorbitals)
        self.assertEqual(qubit, bk_operator)

    def test_jw(self):
        """Check output from Jordan-Wigner transformation."""
        jw_operator = QubitOperator(((1, "Y"), (2, "X")), -0.25j)
        jw_operator += QubitOperator(((1, "Y"), (2, "Y")), -0.25)
        jw_operator += QubitOperator(((1, "X"), (2, "X")), -0.25)
        jw_operator += QubitOperator(((1, "X"), (2, "Y")), 0.25j)
        jw_operator += QubitOperator(((0, "Y"), (1, "Z"), (2, "Z"), (3, "X")), -0.125j)
        jw_operator += QubitOperator(((0, "Y"), (1, "Z"), (2, "Z"), (3, "Y")), 0.125)
        jw_operator += QubitOperator(((0, "X"), (1, "Z"), (2, "Z"), (3, "X")), 0.125)
        jw_operator += QubitOperator(((0, "X"), (1, "Z"), (2, "Z"), (3, "Y")), 0.125j)

        fermion = FermionOperator(((1, 0), (2, 1)), 1.0) + FermionOperator(((0, 1), (3, 0)), 0.5)
        qubit = fermion_to_qubit_mapping(fermion, mapping="JW")
        self.assertEqual(qubit, jw_operator)

    def test_scbk(self):
        """Check output from symmetry-conserving Bravyi-Kitaev transformation.
        """
        scbk_operator = QubitOperator(((0, "Y"),), 1.j)

        fermion = FermionOperator(((2, 0), (0, 1)), 1.) + FermionOperator(((0, 0), (2, 1)), -1.)

        qubit = fermion_to_qubit_mapping(fermion, mapping="SCBK", n_spinorbitals=4, n_electrons=2)
        self.assertEqual(qubit, scbk_operator)

    def test_jkmn(self):
        """Check output from JKMN transform"""
        jkmn_operator = QubitOperator(((0, "X"), (2, "Z"), (3, "X")), 0.125)
        jkmn_operator += QubitOperator(((0, "X"), (2, "Z"), (3, "Y")), 0.125j)
        jkmn_operator += QubitOperator(((0, "Y"), (1, "Z"), (3, "X")), -0.125j)
        jkmn_operator += QubitOperator(((0, "Y"), (1, "Z"), (3, "Y")), 0.125)
        jkmn_operator += QubitOperator(((0, "Z"), (1, "X"), (2, "X")), 0.25j)
        jkmn_operator += QubitOperator(((0, "Z"), (1, "X"), (2, "Y")), 0.25)
        jkmn_operator += QubitOperator(((0, "Z"), (1, "Y"), (2, "X")), -0.25)
        jkmn_operator += QubitOperator(((0, "Z"), (1, "Y"), (2, "Y")), 0.25j)

        fermion = FermionOperator(((1, 0), (2, 1)), 1.0) + FermionOperator(((0, 1), (3, 0)), 0.5)

        qubit = fermion_to_qubit_mapping(fermion, mapping="JKMN", n_spinorbitals=4)
        self.assertEqual(qubit, jkmn_operator)

    def test_scbk_invalid(self):
        """Check if fermion operator fails to conserve number parity or spin
        parity. In either case, scBK is not an appropriate mapping.
        """
        # excitation violating number and spin parity
        fermion = FermionOperator((1, 1), 1.0)
        with self.assertRaises(ValueError):
            fermion_to_qubit_mapping(fermion, mapping="SCBK", n_spinorbitals=2, n_electrons=1)
        # excitation violating spin parity
        fermion = FermionOperator(((0, 1), (1, 0)), 1.0)
        with self.assertRaises(ValueError):
            fermion_to_qubit_mapping(fermion, mapping="SCBK", n_spinorbitals=2, n_electrons=1)

    def test_handle_invalid_mapping(self):
        """Test that error is handled if invalid mapping is requested."""
        fermion = FermionOperator(((1, 0), (2, 1)), 1.0) + FermionOperator(((0, 1), (3, 0)), 0.5)
        with self.assertRaises(ValueError):
            fermion_to_qubit_mapping(fermion, mapping="invalid_mapping")

    def test_eigen(self):
        """Test that all encodings of the operator have the same ground state
        energy.
        """
        fermion = FermionOperator(((1, 0), (3, 1)), 1.0) + FermionOperator(((3, 0), (1, 1)), -1.0)
        ground = np.imag(eigenspectrum(fermion)).min()

        jw_operator = fermion_to_qubit_mapping(fermion, mapping="JW")
        bk_operator = fermion_to_qubit_mapping(fermion, mapping="BK", n_spinorbitals=4)
        scbk_operator = fermion_to_qubit_mapping(fermion, mapping="SCBK", n_spinorbitals=4, n_electrons=2)
        jkmn_operator = fermion_to_qubit_mapping(fermion, mapping="JKMN", n_spinorbitals=4)

        jw_ground = np.linalg.eigvalsh(qubit_operator_sparse(jw_operator).todense()).min()
        bk_ground = np.linalg.eigvalsh(qubit_operator_sparse(bk_operator, n_qubits=4).todense()).min()
        scbk_ground = np.linalg.eigvalsh(qubit_operator_sparse(scbk_operator, n_qubits=2).todense()).min()
        jkmn_ground = np.linalg.eigvalsh(qubit_operator_sparse(jkmn_operator, n_qubits=4).todense()).min()

        self.assertAlmostEqual(ground, jw_ground, places=15)
        self.assertAlmostEqual(ground, bk_ground, places=15)
        self.assertAlmostEqual(ground, scbk_ground, places=15)
        self.assertAlmostEqual(ground, jkmn_ground, places=15)

    def test_spin_order(self):
        """Test that re-ordering of spin-orbitals from alternating up down to
        all up, then all down produces expected result.
        """
        input_operator = FermionOperator(((1, 1), (3, 0), (4, 1), (6, 0)), 1.0) \
                         + FermionOperator(((2, 0), (5, 0), (5, 1), (7, 0)), 1.0)
        expected = FermionOperator(((4, 1), (5, 0), (2, 1), (3, 0)), 1.0) \
                   + FermionOperator(((1, 0), (6, 0), (6, 1), (7, 0)), 1.0)
        reordered_operator = make_up_then_down(input_operator, 8)
        self.assertEqual(expected, reordered_operator)

    def test_eigen_spin_reorder(self):
        """Test that all encodings of the operator have the same ground state
        energy, after re-ordering spins. Skip scBK here as this transformation
        forces spin-reordering.
        """
        fermion = FermionOperator(((1, 0), (3, 1)), 1.0) + FermionOperator(((3, 0), (1, 1)), -1.0)
        ground = np.imag(eigenspectrum(fermion)).min()

        jw_operator = fermion_to_qubit_mapping(fermion, mapping="JW", n_spinorbitals=4, up_then_down=True)
        bk_operator = fermion_to_qubit_mapping(fermion, mapping="BK", n_spinorbitals=4, up_then_down=True)
        jkmn_operator = fermion_to_qubit_mapping(fermion, mapping="JKMN", n_spinorbitals=4, up_then_down=True)

        jw_ground = np.linalg.eigvalsh(qubit_operator_sparse(jw_operator).todense()).min()
        bk_ground = np.linalg.eigvalsh(qubit_operator_sparse(bk_operator, n_qubits=4).todense()).min()
        jkmn_ground = np.linalg.eigvalsh(qubit_operator_sparse(jkmn_operator, n_qubits=4).todense()).min()

        self.assertEqual(ground, jw_ground)
        self.assertEqual(ground, bk_ground)
        self.assertEqual(ground, jkmn_ground)

    def test_scbk_reorder(self):
        """scBK forces spin-orbital ordering to all up then all down. Check that
        the qubit Hamiltonian returned is the same whether the user passes a
        FermionOperator with this ordering, or not.
        """
        fermion = FermionOperator(((2, 0), (0, 1)), 1.) + FermionOperator(((0, 0), (2, 1)), -1.)
        scBK_reordered = fermion_to_qubit_mapping(fermion_operator=fermion,
                                                  mapping="scBK",
                                                  n_spinorbitals=4,
                                                  n_electrons=2,
                                                  up_then_down=True)
        scBK_notreordered = fermion_to_qubit_mapping(fermion_operator=fermion,
                                                     mapping="scBK",
                                                     n_spinorbitals=4,
                                                     n_electrons=2,
                                                     up_then_down=False)
        self.assertEqual(scBK_reordered, scBK_notreordered)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from openfermion import get_sparse_operator, linalg

from tangelo.linq import Simulator, Circuit
from tangelo.molecule_library import mol_H4_sto3g
from tangelo.toolboxes.circuits import get_orbital_rotations

# Initiate simulator using cirq as it has the same ordering as openfermion and we are using an exact eigenvector to test
sim = Simulator(target="cirq")


class diagonal_coulomb_Test(unittest.TestCase):

    def test_orbital_rotations(self):
        """Test calculating energy expectation value of H4 hamiltonian by decomposing into diagional terms"""

        # Generate ground state wavefunction from JW transformed fermionic_hamiltonian
        ham = get_sparse_operator(mol_H4_sto3g.fermionic_hamiltonian)
        eig, state_vec = linalg.get_ground_state(ham)

        # Generate necessary circuits and operators
        orb_rots = get_orbital_rotations(mol_H4_sto3g)

        # Run each set of gates that diagonalizes the set of terms and calculate energy
        energy = 0
        for i in range(orb_rots.n_rotations):
            # Get appropriate qubit operator to measure after each rotation.
            qu_op = orb_rots.qubit_operators[i]
            # Rotate state and measure all terms using one set of measurements.
            freqs, _ = sim.simulate(Circuit(orb_rots.rotation_gates[i]), initial_statevector=state_vec)
            for term, coeff in qu_op.terms.items():
                energy += coeff*sim.get_expectation_value_from_frequencies_oneterm(term, frequencies=freqs)

        self.assertAlmostEqual(energy, eig, places=6)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from scipy import stats
from collections import Counter


def get_resampled_frequencies(freq_dict, ncount):
    """From a frequencies dictionary, makes a set of samples consistent with
    those frequencies and resample from that set to obtain a new frequencies
    dictionary consistent with ncount measurements.

    Args:
        freq_dict (array): dictionary of measurement/sample frequencies.
        ncount (int): number of shots/samples to generate resampled frequencies.

    Returns
        dict: new frequencies dictionary with resampled distribution.
    """

    length_dict = len(freq_dict.keys())
    xk = np.empty(length_dict, dtype=np.int64)
    pk = np.empty(length_dict, dtype=float)

    # Convert to arrays of integers and frequencies and create generator for samples
    for i, (k, v) in enumerate(freq_dict.items()):
        xk[i] = int(k, 2)
        pk[i] = v
    distr = stats.rv_discrete(name="distr", values=(xk, pk))

    # Obtain output bitstring string format
    n_qubits = len(list(freq_dict.keys())[0])
    format_specifier = "0"+str(n_qubits)+"b"

    # Generate samples from distribution. Cut in chunks to ensure samples fit in memory, gradually accumulate
    chunk_size = 10**7
    n_chunks = ncount // chunk_size
    freqs_shots = Counter()

    for i in range(n_chunks+1):
        this_chunk = ncount % chunk_size if i == n_chunks else chunk_size
        samples = distr.rvs(size=this_chunk)
        freqs_shots += Counter(samples)
    frequencies = {format(k, format_specifier): v / ncount for k, v in freqs_shots.items()}

    return frequencies



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions helping with quantum circuit format conversion between abstract
format and projectq format

In order to produce an equivalent circuit for the target backend, it is
necessary to account for:
- how the gate names differ between the source backend to the target backend.
- how the order and conventions for some of the inputs to the gate operations
    may also differ.
"""

import re
from tangelo.linq import Gate, Circuit


def get_projectq_gates():
    """Map gate name of the abstract format to the equivalent gate name used in
    projectq API and supported gates:
    https://projectq.readthedocs.io/en/latest/projectq.ops.html
    """

    GATE_PROJECTQ = dict()
    for name in {"H", "X", "Y", "Z", "S", "T"}:
        GATE_PROJECTQ[name] = name
    for name in {"RX", "RY", "RZ", "MEASURE"}:
        GATE_PROJECTQ[name] = name[0] + name[1:].lower()
    GATE_PROJECTQ["CNOT"] = "CX"
    GATE_PROJECTQ["PHASE"] = "R"

    return GATE_PROJECTQ


def translate_projectq(source_circuit):
    """Take in an abstract circuit, return a string containing equivalent
    projectq instructions.

    Args:
        source_circuit: quantum circuit in the abstract format.

    Returns:
        str: the corresponding projectq instructions (allocation , gates,
            deallocation).
    """

    GATE_PROJECTQ = get_projectq_gates()

    projectq_circuit = ""
    for i in range(source_circuit.width):
        projectq_circuit += f"Allocate | Qureg[{i}]\n"

    for gate in source_circuit._gates:
        if gate.name in {"H", "X", "Y", "Z", "S", "T", "MEASURE"}:
            projectq_circuit += f"{GATE_PROJECTQ[gate.name]} | Qureg[{gate.target[0]}]\n"
        elif gate.name in {"RX", "RY", "RZ", "PHASE"}:
            projectq_circuit += f"{GATE_PROJECTQ[gate.name]}({gate.parameter}) | Qureg[{gate.target[0]}]\n"
        elif gate.name in {"CNOT"}:
            projectq_circuit += f"{GATE_PROJECTQ[gate.name]} | ( Qureg[{gate.control[0]}], Qureg[{gate.target[0]}] )\n"
        else:
            raise ValueError(f"Gate '{gate.name}' not supported on backend projectQ")

    return projectq_circuit


def _translate_projectq2abs(projectq_str):
    """Convenience function to help people move away from their ProjectQ
    workflow. Take ProjectQ instructions, expressed as a string, similar to one
    from the ProjectQ `CommandPrinter` engine. Will bundle all qubit allocation
    (resp. deallocation) at the beginning (resp. end) of the circuit. Example
    available in the `examples` folder.

    Args:
        projectq_str(str): ProjectQ program, as a string (Allocate, Deallocate,
            gate operations...).

    Returns:
        Circuit: corresponding quantum circuit in the abstract format.
    """

    # Get dictionary of gate mapping, as the reverse dictionary of abs -> projectq translation
    GATE_PROJECTQ = get_projectq_gates()
    gate_mapping = {v: k for k, v in GATE_PROJECTQ.items()}

    # TODO account for mid-circuit measurements, only ignore final measurements
    # Ignore Measure instructions
    projectq_str = re.sub(r'Measure(.*)\n', '', projectq_str)

    # Ignore allocate and deallocate instructions.
    # Number of qubits is inferred by the abstract circuit, no (de)allocation will occur mid-circuit.
    projectq_str = re.sub(r'(.*)llocate(.*)\n', '', projectq_str)
    projectq_gates = [instruction for instruction in projectq_str.split("\n") if instruction]

    # Translate instructions to abstract gates
    abs_circ = Circuit()
    for projectq_gate in projectq_gates:

        # Extract gate name, qubit indices and parameter value (single parameter for now)
        gate_name = re.split(' \| |\(', projectq_gate)[0]
        qubit_indices = [int(index) for index in re.findall('Qureg\[(\d+)\]', projectq_gate)]
        parameters = [float(index) for index in re.findall('\((.*)\)', projectq_gate) if "Qureg" not in index]

        if gate_name in {"H", "X", "Y", "Z", "S", "T"}:
            gate = Gate(gate_mapping[gate_name], qubit_indices[0])
        elif gate_name in {"Rx", "Ry", "Rz", "PHASE"}:
            gate = Gate(gate_mapping[gate_name], qubit_indices[0], parameter=parameters[0])
        # #TODO: Rethink the use of enums for gates to set the equality CX=CNOT and enable other refactoring
        elif gate_name in {"CX"}:
            gate = Gate(gate_mapping[gate_name], qubit_indices[1], control=qubit_indices[0])
        else:
            raise ValueError(f"Gate '{gate_name}' not supported with project2abs translation")
        abs_circ.add_gate(gate)

    return abs_circ



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions on QubitOperators, such as exporting Openfermion
QubitOperator to file and vice-versa, or to visualize its coefficients for
example.
"""

import math
import numpy as np


def print_histogram_coeffs(qb_ham):
    """Convenience function printing a matplotlib histogram of the magnitudes
    of the coefficient in a QubitOperator object. Combine with the compress
    method of the QubitOperator class, this allows users to quickly identify
    what terms in the operator can be discarded, depending on the target
    accuracy of calculations.
    """

    import matplotlib
    matplotlib.use('TkAgg')
    import matplotlib.pyplot as plt

    coefs = [max(abs(coef.real), abs(coef.imag)) for _, coef in qb_ham.terms.items()]
    magn = [math.floor(math.log10(coef)) for coef in coefs if coef != 0.]
    magn = np.array(magn)
    bins = list(np.arange(magn.min(), magn.max() + 2))

    plt.hist(magn, bins=bins, rwidth=.3)
    plt.xticks(ticks=bins, labels=[f"1e{i}" for i in bins])
    plt.grid(True)
    plt.xlabel('Magnitude')
    plt.title(f'Magnitude of qubit operator coeffs (Total = {len(coefs)})')
    plt.show()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import numpy as np

from tangelo.toolboxes.operators import MultiformOperator
from tangelo.toolboxes.operators.z2_tapering import get_clifford_operators, get_unitary, get_eigenvalues


class Z2TaperingHelperFunctionsTest(unittest.TestCase):

    def test_get_clifford_operators(self):
        """Test get_clifford_operators function with a given kernel."""

        kernel = np.array([
            [0, 0, 0, 0, 1, 0, 0, 1],
            [0, 0, 0, 0, 1, 1, 0, 0],
            [0, 0, 0, 0, 1, 0, 1, 0]
        ])
        clifford_ops, indices = get_clifford_operators(kernel)

        self.assertEqual(3, len(clifford_ops))

        np.testing.assert_array_equal(np.array([3, 1, 2]), indices)

        ref_clif1 = {((3, 'X'),): np.sqrt(0.5), ((0, 'Z'), (3, 'Z')): np.sqrt(0.5)}
        self.assertDictEqual(ref_clif1, clifford_ops[0].terms)

        ref_clif2 = {((1, 'X'),): np.sqrt(0.5), ((0, 'Z'), (1, 'Z')): np.sqrt(0.5)}
        self.assertDictEqual(ref_clif2, clifford_ops[1].terms)

        ref_clif3 = {((2, 'X'),): np.sqrt(0.5), ((0, 'Z'), (2, 'Z')): np.sqrt(0.5)}
        self.assertDictEqual(ref_clif3, clifford_ops[2].terms)

    def test_get_unitary(self):
        """Test get_unitary function with a given set of Clifford operators."""

        ref_u_terms = {
            ((1, 'X'), (2, 'X'), (3, 'X')): (0.35355339059327384+0j),
            ((0, 'Z'), (1, 'X'), (2, 'Z'), (3, 'X')): (0.35355339059327384+0j),
            ((0, 'Z'), (1, 'Z'), (2, 'X'), (3, 'X')): (0.35355339059327384+0j),
            ((1, 'Z'), (2, 'Z'), (3, 'X')): (0.35355339059327384+0j),
            ((0, 'Z'), (1, 'X'), (2, 'X'), (3, 'Z')): (0.35355339059327384+0j),
            ((1, 'X'), (2, 'Z'), (3, 'Z')): (0.35355339059327384+0j),
            ((1, 'Z'), (2, 'X'), (3, 'Z')): (0.35355339059327384+0j),
            ((0, 'Z'), (1, 'Z'), (2, 'Z'), (3, 'Z')): (0.35355339059327384+0j)
        }

        factors = np.array([np.sqrt(0.5), np.sqrt(0.5)])
        clif1 = np.array([
            [1, 0, 0, 1],
            [0, 0, 0, 2]
        ])
        clif2 = np.array([
            [1, 1, 0, 0],
            [0, 2, 0, 0]
        ])
        clif3 = np.array([
            [1, 0, 1, 0],
            [0, 0, 2, 0]
        ])

        cliff_ops = [
            MultiformOperator.from_integerop(clif1, factors),
            MultiformOperator.from_integerop(clif2, factors),
            MultiformOperator.from_integerop(clif3, factors),
        ]

        U = get_unitary(cliff_ops)
        self.assertDictEqual(ref_u_terms, U.terms)

    def test_get_eigenvalues(self):
        """Test get_eigenvalues function with a given kernel operator
        representing symmetries.
        """

        integers = np.array([
            [1, 0, 0, 1],
            [1, 1, 0, 0],
            [1, 0, 1, 0]
        ])
        kernel_op = MultiformOperator.from_integerop(integers, np.ones(integers.shape[0]))

        np.testing.assert_array_equal([-1, 1, -1], get_eigenvalues(kernel_op.binary, 4, 2, 0, "JW", False))
        np.testing.assert_array_equal([-1, -1, 1], get_eigenvalues(kernel_op.binary, 4, 2, 0, "JW", True))
        np.testing.assert_array_equal([-1, -1, -1], get_eigenvalues(kernel_op.binary, 4, 2, 0, "BK", False))
        np.testing.assert_array_equal([-1, 1, 1], get_eigenvalues(kernel_op.binary, 4, 2, 0, "BK", True))
        np.testing.assert_array_equal([1, -1, 1], get_eigenvalues(kernel_op.binary, 4, 2, 0, "JKMN", False))
        np.testing.assert_array_equal([1, 1, -1], get_eigenvalues(kernel_op.binary, 4, 2, 0, "JKMN", True))


if __name__ == "__main__":
    unittest.main()
